from tensorflow import keras
import tensorflow as tf
from warnings import warn
from teras.utils import sparsemoid
from teras.activations import sparsemax
import tensorflow_probability as tfp
from teras.utils import get_activation


@keras.saving.register_keras_serializable(package="teras.layers.node")
class ObliviousDecisionTree(keras.layers.Layer):
    """
        Oblivious Decision Tree layer as proposed by Sergei Popov et al.
        in paper Neural Oblivious Decision Ensembles
        for Deep Learning on Tabular Data

        Reference(s):
            https://arxiv.org/abs/1909.06312

        Args:
       num_trees: ``int``, default 128,
            Number of trees

       depth: ``int``, default 6,
           Number of splits in every tree

       tree_dim: ``int``, default 1,
           Number of response channels in the response of individual tree

       max_features: ``int``,
           Maximum number of features to use. If None, all features in the input dataset will be used.

       input_dropout: ``float``, default 0.,
           Dropout rate to apply to inputs.

       choice_function:
           f(tensor, dim) -> R_simplex computes feature weights s.t. f(tensor, dim).sum(dim) == 1
           By default, sparsemax is used.

       bin_function:
           f(tensor) -> R[0, 1], computes tree leaf weights
           By default, `sparsemoid` is used.

       response_initializer: default "random_normal",
           Initializer for tree output tensor. Any format that is acceptable by the keras initializers.

       selection_logits_initializer: default "random_uniform",
           Initializer for logits that select features for the tree
           Both thresholds and scales are initialized with data-aware initialization function.

       threshold_init_beta: ``float``, default 1.0,
           Initializes threshold to a q-th quantile of data points
           where q ~ Beta(:threshold_init_beta:, :threshold_init_beta:)
           If this param is set to 1, initial thresholds will have the same distribution as data points
           If greater than 1 (e.g. 10), thresholds will be closer to median data value
           If less than 1 (e.g. 0.1), thresholds will approach min/max data values.

       threshold_init_cutoff: ``float``, default 1.0,
           Threshold log-temperatures initializer, \in (0, inf)
           By default(1.0), log-temperatures are initialized in such a way that all bin selectors
           end up in the linear region of sparse-sigmoid. The temperatures are then scaled by this parameter.
           Setting this value > 1.0 will result in some margin between data points and sparse-sigmoid cutoff value
           Setting this value < 1.0 will cause (1 - value) part of data points to end up in flat sparse-sigmoid region
           For instance, threshold_init_cutoff = 0.9 will set 10% points equal to 0.0 or 1.0
           Setting this value > 1.0 will result in a margin between data points and sparse-sigmoid cutoff value
           All points will be between (0.5 - 0.5 / threshold_init_cutoff) and (0.5 + 0.5 / threshold_init_cutoff)
    """
    def __init__(self,
                 num_trees: int = 128,
                 depth: int = 6,
                 tree_dim: int = 1,
                 choice_function=None,
                 bin_function=None,
                 response_initializer="random_normal",
                 selection_logits_initializer="random_uniform",
                 threshold_init_beta: float = 1.0,
                 threshold_init_cutoff: float = 1.0,
                 **kwargs):
        super().__init__(**kwargs)
        self.num_trees = num_trees
        self.depth = depth
        self.tree_dim = tree_dim
        self.choice_function = sparsemax if choice_function is None else get_activation(choice_function)
        self.bin_function = sparsemoid if bin_function is None else get_activation(bin_function)
        self.response_initializer = response_initializer
        self.selection_logits_initializer = selection_logits_initializer
        self.threshold_init_beta = threshold_init_beta
        self.threshold_init_cutoff = threshold_init_cutoff
        self.initialized = False

    def build(self, input_shape):
        input_dim = input_shape[-1]
        self.response = self.add_weight(initializer=self.response_initializer,
                                        shape=(self.num_trees, self.tree_dim, 2 ** self.depth))
        self.feature_selection_logits = self.add_weight(initializer=self.selection_logits_initializer,
                                                        shape=(input_dim, self.num_trees, self.depth))
        self.feature_thresholds = tf.Variable(initial_value=keras.initializers.zeros()(
                                                                                shape=(self.num_trees, self.depth),
                                                                                dtype="float32"),
                                              shape=[self.num_trees, self.depth])

        self.log_temperatures = tf.Variable(initial_value=keras.initializers.zeros()(
                                                                                shape=(self.num_trees, self.depth),
                                                                                dtype="float32"),
                                            shape=[self.num_trees, self.depth])

        indices = keras.backend.arange(2 ** self.depth)
        offsets = 2 ** keras.backend.arange(self.depth)
        bin_codes = tf.reshape(indices, shape=(1, -1)) // tf.reshape(offsets, shape=(-1, 1)) % 2
        bin_codes = tf.cast(bin_codes, dtype="float32")
        bin_codes_1hot = tf.stack([bin_codes, 1.0 - bin_codes], axis=-1)
        self.bin_codes_1hot = tf.Variable(initial_value=bin_codes_1hot, trainable=False)

    def _data_aware_initialization(self, inputs, eps=1e-6):
        """
        Data aware initialization as proposed by the paper authors.
        As the name implies, the function initializes variables based on the inputs.
        Hence, it is called when the layer's call method is called on some inputs
        and not in __init__ or build methods.
        """
        assert len(inputs.shape) == 2
        if tf.shape(inputs)[0] < 1000:
            warn("Data-aware initialization is performed on less than 1000 data points. This may cause instability."
                 "To avoid potential problems, run this model on a data batch with at least 1000 data samples."
                 "You can do so manually before training.")
        feature_selectors = self.choice_function(self.feature_selection_logits, axis=0)
        feature_values = tf.einsum('bi,ind->bnd', inputs, feature_selectors)
        beta_dist = tfp.distributions.Beta(self.threshold_init_beta, self.threshold_init_beta)
        percentiles_q = 100 * beta_dist.sample([self.num_trees * self.depth])

        flattened_feature_values = tf.map_fn(tf.keras.backend.flatten, feature_values)

        feature_thresholds = tfp.stats.percentile(flattened_feature_values, percentiles_q)
        feature_thresholds = tf.reshape(feature_thresholds, shape=(self.num_trees, self.depth))
        temperatures = tfp.stats.percentile(tf.abs(feature_values - feature_thresholds),
                                     q=100 * min(1.0, self.threshold_init_cutoff), axis=0)

        # if threshold_init_cutoff > 1, scale everything down by it
        temperatures /= max(1.0, self.threshold_init_cutoff)
        log_tempratures = tf.math.log(temperatures + eps)
        log_tempratures = tf.reshape(log_tempratures, shape=[self.num_trees, self.depth])

        self.feature_thresholds.assign(feature_thresholds)
        self.log_temperatures.assign(log_tempratures)

    def call(self, inputs):
        if len(inputs.shape) > 2:
            return self.call(tf.reshape(tf.reshape(inputs, shape=(-1, inputs.shape[-1])), shape=(*inputs.shape[:-1], -1)))
        if not self.initialized:
            self._data_aware_initialization(inputs)
            self.initialized = True

        feature_logits = self.feature_selection_logits
        features_selectors = self.choice_function(feature_logits, axis=0)
        feature_values = tf.einsum('bi,ind->bnd', inputs, features_selectors)
        threshold_logits = (feature_values - self.feature_thresholds) * tf.exp(-self.log_temperatures)
        threshold_logits = tf.stack([-threshold_logits, threshold_logits], axis=-1)
        bins = self.bin_function(threshold_logits)
        bin_matches = tf.einsum('btds,dcs->btdc', bins, self.bin_codes_1hot)
        response_weights = tf.reduce_prod(bin_matches, axis=-2)
        response = tf.einsum('bnd,ncd->bnc', response_weights, self.response)
        return tf.map_fn(keras.backend.flatten, response)

    def get_config(self):
        config = super().get_config()
        new_config = {'num_trees': self.num_trees,
                      'depth': self.depth,
                      'tree_dim': self.tree_dim,
                      'choice_function': self.choice_function,
                      'bin_function': self.bin_function,
                      'response_initializer': self.response_initializer,
                      'selection_logits_initializer': self.selection_logits_initializer,
                      'threshold_init_beta': self.threshold_init_beta,
                      'threshold_init_cutoff': self.threshold_init_cutoff,
                      }

        config.update(new_config)
        return config


import tensorflow as tf
from teras.layers.vime.vime_encoder import VimeEncoder


def test_vime_feature_estimator_valid_call():
    inputs = tf.ones((8, 5))
    encoder = VimeEncoder(units=16)
    outputs = encoder(inputs)


def test_vime_feature_estimator_output_shape():
    inputs = tf.ones((8, 5))
    encoder = VimeEncoder(units=16)
    outputs = encoder(inputs)
    assert len(tf.shape(outputs)) == 2
    assert tf.shape(outputs)[0] == 8
    assert tf.shape(outputs)[1] == 16



import tensorflow as tf
from tensorflow.keras import losses


def elbo_loss_tvae(data_dim,
                   real_samples=None,
                   generated_samples=None,
                   meta_data=None,
                   sigmas=None,
                   mean=None,
                   log_var=None,
                   loss_factor=2):
    """
    Evidence Lower Bound (ELBO) Loss [1] adapted for
    TVAE architecture [2] proposed by
    Lei Xu et al. in the paper,
    "Modeling Tabular data using Conditional GAN".

    Reference(s):
        [1]: https://arxiv.org/abs/1312.6114
        [2]: https://arxiv.org/abs/1907.00503

    Args:
        data_dim: Dimensionality of the input dataset.
        real_samples: Samples drawn from the input dataset.
        generated_samples: Samples generated by the decoder.
        meta_data: A namedtuple containing meta data for all
            features in the input data. This meta data is computed
            during the data transformation step and can be accessed
            from `.get_meta_data()` method of DataTransformer instance.
        sigmas: Sigma values returned by decoder
        mean: Mean values returned by encoder
        log_var: Log var values returned by encoder
        loss_factor: Loss factor that controls how much the cross entropy
            loss contributes to the overall loss.
            It is directly proportional to the cross entropy loss.

    Returns:
        Elbo loss adapted for the TVAE model.
    """
    loss = []
    cross_entropy = losses.SparseCategoricalCrossentropy(from_logits=True,
                                                         reduction=losses.Reduction.SUM)
    cont_i = 0  # numerical index
    cat_i = 0  # categorical index
    for i, relative_index in enumerate(meta_data.relative_indices_all):
        # the first k features are numerical
        # TODO: replace it with len(meta_data.numerical_features) but it will require changing the DataTransformer
        if i < len(meta_data.numerical.num_valid_clusters_all):
            # each numerical feature is of the form
            # [alpha, beta1, beta2...beta(n)] where n is the number of clusters

            # calculate alpha loss
            std = sigmas[relative_index]
            eq = real_samples[:, relative_index] - tf.nn.tanh(generated_samples[:, relative_index])
            loss_temp = tf.reduce_sum((eq ** 2 / 2 / (std ** 2)))
            loss.append(loss_temp)
            loss_temp = tf.math.log(std) * tf.cast(tf.shape(real_samples)[0], dtype=tf.float32)
            loss.append(loss_temp)

            # calculate betas loss
            num_clusters = meta_data.numerical.num_valid_clusters_all[cont_i]
            logits = generated_samples[:, relative_index + 1: relative_index + 1 + num_clusters]
            labels = tf.argmax(real_samples[:, relative_index + 1: relative_index + 1 + num_clusters], axis=-1)
            cross_entropy_loss = cross_entropy(y_pred=logits, y_true=labels)
            loss.append(cross_entropy_loss)
            cont_i += 1
        else:
            num_categories = meta_data.categorical.num_categories_all[cat_i]
            logits = generated_samples[:, relative_index: relative_index + num_categories]
            labels = tf.argmax(real_samples[:, relative_index: relative_index + num_categories], axis=-1)
            cross_entropy_loss = cross_entropy(y_pred=logits, y_true=labels)
            loss.append(cross_entropy_loss)
            cat_i += 1
    KLD = -0.5 * tf.reduce_sum(1 + log_var - mean ** 2 - tf.exp(log_var))
    loss_1 = tf.reduce_sum(loss) * loss_factor / data_dim
    loss_2 = KLD / data_dim
    final_loss = loss_1 + loss_2
    return final_loss



import tensorflow as tf
from teras.layers.tabtransformer.tabtransformer_column_embedding import TabTransformerColumnEmbedding


def test_tabtransformer_column_embedding_output_shape():
    inputs = tf.ones((16, 8, 32))
    ce = TabTransformerColumnEmbedding(num_categorical_features=8,
                                       embedding_dim=32)
    outputs = ce(inputs)
    assert len(tf.shape(outputs)) == 3
    assert tf.shape(outputs)[0] == 16
    assert tf.shape(outputs)[1] == 8
    assert tf.shape(outputs)[2] == 32




import tensorflow as tf
from tensorflow import keras
from teras.layerflow.models.tabnet import (TabNet,
                                           TabNetPretrainer)
from teras.layers import (CategoricalFeatureEmbedding,
                          TabNetEncoder,
                          TabNetDecoder)
from teras.utils import get_features_metadata_for_embedding
import pandas as pd
import numpy as np
import os
import tensorflow_probability as tfp


class TabNetTest(tf.test.TestCase):
    def setUp(self):
        data = pd.DataFrame({"income": np.ones(10),
                             "goals": np.ones(10),
                             'player_level': [5, 7, 9, 8, 9, 10, 9, 7, 8, 9],
                             'shirt_number': [7, 10, 10, 7, 7, 10, 10, 10, 7, 10]})
        categorical_feats = ["player_level", "shirt_number"]
        numerical_feats = ["income", "goals"]
        self.features_metadata = get_features_metadata_for_embedding(pd.DataFrame(data),
                                                                     categorical_features=categorical_feats,
                                                                     numerical_features=numerical_feats)
        self.categorical_feature_embedding = CategoricalFeatureEmbedding(features_metadata=self.features_metadata,
                                                                         embedding_dim=1)
        self.encoder = TabNetEncoder(data_dim=4)
        self.head = keras.layers.Dense(1)
        self.data_batch = data.values

    def test_valid_call(self):
        model = TabNet(input_dim=4,
                       features_metadata=self.features_metadata,
                       categorical_feature_embedding=self.categorical_feature_embedding,
                       encoder=self.encoder,
                       head=self.head)
        model(self.data_batch)

    def test_save_and_load(self):
        model = TabNet(input_dim=4,
                       features_metadata=self.features_metadata,
                       categorical_feature_embedding=self.categorical_feature_embedding,
                       encoder=self.encoder,
                       head=self.head)
        save_path = os.path.join(self.get_temp_dir(), "tabnet_lf.keras")
        model.save(save_path, save_format="keras_v3")
        reloaded_model = keras.models.load_model(save_path)
        outputs_original = model(self.data_batch)
        outputs_reloaded = reloaded_model(self.data_batch)
        self.assertAllClose(outputs_original, outputs_reloaded)


class TabNetPretrainerTest(tf.test.TestCase):
    def setUp(self):
        data = pd.DataFrame({"income": np.ones(10),
                             "goals": np.ones(10),
                             'player_level': [5, 7, 9, 8, 9, 10, 9, 7, 8, 9],
                             'shirt_number': [7, 10, 10, 7, 7, 10, 10, 10, 7, 10]})
        categorical_feats = ["player_level", "shirt_number"]
        numerical_feats = ["income", "goals"]
        self.features_metadata = get_features_metadata_for_embedding(pd.DataFrame(data),
                                                                     categorical_features=categorical_feats,
                                                                     numerical_features=numerical_feats)
        self.categorical_feature_embedding = CategoricalFeatureEmbedding(features_metadata=self.features_metadata,
                                                                         embedding_dim=1)
        self.encoder = TabNetEncoder(data_dim=4)
        self.decoder = TabNetDecoder(data_dim=4)
        self.binary_mask_generator = tfp.distributions.Binomial(total_count=1,
                                                                probs=0.3,
                                                                name="binary_mask_generator")
        self.data_batch = data.values
        self.mask = self.binary_mask_generator.sample(self.data_batch.shape)

    def test_valid_call(self):
        base_model = TabNet(input_dim=4,
                            features_metadata=self.features_metadata,
                            categorical_feature_embedding=self.categorical_feature_embedding,
                            encoder=self.encoder)
        pretrainer = TabNetPretrainer(model=base_model,
                                      features_metadata=self.features_metadata,
                                      decoder=self.decoder,
                                      missing_feature_probability=0.3)
        pretrainer(self.data_batch, mask=self.mask)

    def test_save_and_load(self):
        base_model = TabNet(input_dim=4,
                            features_metadata=self.features_metadata,
                            categorical_feature_embedding=self.categorical_feature_embedding,
                            encoder=self.encoder)
        pretrainer = TabNetPretrainer(model=base_model,
                                      features_metadata=self.features_metadata,
                                      decoder=self.decoder,
                                      missing_feature_probability=0.3)
        save_path = os.path.join(self.get_temp_dir(), "tabnet_pretrainer_lf.keras")
        pretrainer.save(save_path, save_format="keras_v3")
        reloaded_model = keras.models.load_model(save_path)
        outputs_original = pretrainer(self.data_batch, mask=self.mask)
        outputs_reloaded = reloaded_model(self.data_batch, mask=self.mask)
        # We can't check for AllClose here.. will investigate why



from tensorflow import keras
from teras.utils.utils import (serialize_layers_collection,
                               deserialize_layers_collection)
from teras.utils.types import LayersList


@keras.saving.register_keras_serializable(package="teras.layerflow.models")
class NODE(keras.Model):
    """
    Neural Oblivious Decision Tree (NODE) model with LayerFlow design.
    It is based on the NODE architecture proposed by Sergei Popov et al.
    in the paper,
    Neural Oblivious Decision Ensembles for Deep Learning on Tabular Data

    Reference(s):
        https://arxiv.org/abs/1909.06312

    Args:
        input_dim: ``int``,
            Dimensionality of the input dataset,
            or, the number of features in the dataset.

        tree_layers: ``List[layers.Layer]``,
            A list or tuple of `ObliviousDecisionTree` layers instances, or any custom
            layer that can work in its place.
            If None, default number of `ObliviousDecisionTree` layers with default values
            will be used.
            You can import the `ObliviousDecisionTree` layer as follows,
                >>> from teras.layers import ObliviousDecisionTree

        feature_selector: ``keras.layers.Layer``:
            An instance of ``NodeFeatureSelector`` layer that selects features based on
            In None, all features will be used (no ``NodeFeatureSelector`` layer will be applied).
            You can import the ``NodeFeatureSelector`` layer as follows,
                >>> from teras.layers import NodeFeatureSelector

        dropout: `keras.layers.Layer``,
            An instance of ``Dropout`` layer to apply over inputs.
            If None, no dropout will be applied.
            You can import the ``Dropout`` layer as follows,
                >>> from keras.layers import Dropout

        head: ``keras.layers.Layer``,
            An instance of either ``ClassificationHead`` or ``RegressionHead`` layers,
            depending on the task at hand.
            You can import the ``ClassificationHead`` and ``RegressionHead`` layers as follows,
                >>> from teras.layers import ClassificationHead
                >>> from teras.layers import RegressionHead
    """
    def __init__(self,
                 input_dim: int,
                 tree_layers: LayersList = None,
                 feature_selector: keras.layers.Layer = None,
                 dropout: keras.layers.Layer = None,
                 head: keras.layers.Layer = None,
                 **kwargs):
        if not isinstance(tree_layers, (list, tuple)):
            raise ValueError("`tree_layers` must be a list or tuple of `ObliviousDecisionTree` layers, "
                             f"but received type: {type(tree_layers)}.")

        inputs = keras.layers.Input(shape=(input_dim,),
                                    name="inputs")
        x_out = inputs
        initial_features = input_dim
        for layer in tree_layers:
            x = x_out
            if feature_selector is not None:
                x = feature_selector(x)
            if dropout is not None:
                x = dropout(x)
            h = layer(x)
            x_out = keras.layers.Concatenate(axis=-1)([x_out, h])
        outputs = x_out[..., initial_features:]
        if head is not None:
            outputs = head(outputs)
        super().__init__(inputs=inputs,
                         outputs=outputs,
                         **kwargs)

        self.input_dim = input_dim
        self.tree_layers = tree_layers
        self.head = head
        self.feature_selector = feature_selector
        self.dropout = dropout

    def get_config(self):
        config = super().get_config()
        config.update({'input_dim': self.input_dim,
                       'tree_layers': serialize_layers_collection(self.tree_layers),
                       'feature_selector': keras.layers.serialize(self.feature_selector),
                       'dropout': keras.layers.serialize(self.dropout),
                       'head': keras.layers.serialize(self.head)
                       })
        return config

    @classmethod
    def from_config(cls, config):
        input_dim = config.pop("input_dim")
        tree_layers = deserialize_layers_collection(config.pop("tree_layers"))
        feature_selector = keras.layers.deserialize(config.pop("feature_selector"))
        dropout = keras.layers.deserialize(config.pop("dropout"))
        head = keras.layers.deserialize(config.pop("head"))
        return cls(input_dim=input_dim,
                   tree_layers=tree_layers,
                   feature_selector=feature_selector,
                   dropout=dropout,
                   head=head,
                   **config)



import tensorflow as tf
from teras.layers.saint.multi_head_inter_sample_attention import MultiHeadInterSampleAttention


def test_saint_multi_head_inter_sample_attention_output_shape():
    multihead_attention = MultiHeadInterSampleAttention(key_dim=32)
    inputs = tf.ones(shape=(16, 10, 32))
    outputs = multihead_attention(inputs)
    assert len(tf.shape(outputs)) == 3
    assert tf.shape(outputs)[0] == 16   # number of items in each column
    assert tf.shape(outputs)[1] == 10
    assert tf.shape(outputs)[2] == 32



import tensorflow as tf
from tensorflow import keras
from teras.models.dnfnet import DNFNetClassifier, DNFNetRegressor
import os


class DNFNetClassifierTest(tf.test.TestCase):
    def test_valid_call(self):
        inputs = tf.ones((8, 5))
        model = DNFNetClassifier(num_classes=2,
                                 num_formulas=8,
                                 input_dim=5)
        model(inputs)

    def test_save_and_load(self):
        inputs = tf.ones((8, 5))
        model = DNFNetClassifier(num_classes=2,
                                 num_formulas=8,
                                 input_dim=5)
        save_path = os.path.join(self.get_temp_dir(), "dnfnet_classifier.keras")
        model.save(save_path, save_format="keras_v3")
        reloaded_model = keras.models.load_model(save_path)
        outputs_original = model(inputs)
        outputs_reloaded = reloaded_model(inputs)
        self.assertAllClose(outputs_original, outputs_reloaded)
        self.assertAllClose(model.weights, reloaded_model.weights)


class DNFNetRegressorTest(tf.test.TestCase):
    def test_valid_call(self):
        inputs = tf.ones((8, 5))
        model = DNFNetRegressor(num_outputs=2,
                                num_formulas=8,
                                input_dim=5)
        model(inputs)

    def test_save_and_load(self):
        inputs = tf.ones((8, 5))
        model = DNFNetRegressor(num_outputs=2,
                                num_formulas=8,
                                input_dim=5)
        save_path = os.path.join(self.get_temp_dir(), "dnfnet_regressor.keras")
        model.save(save_path, save_format="keras_v3")
        reloaded_model = keras.models.load_model(save_path)
        outputs_original = model(inputs)
        outputs_reloaded = reloaded_model(inputs)
        self.assertAllClose(outputs_original, outputs_reloaded)
        self.assertAllClose(model.weights, reloaded_model.weights)



from teras.config.base import BaseConfig


class TabTransformerConfig(BaseConfig):
    embedding_dim: int = 32
    num_transformer_layers: int = 6
    num_attention_heads: int = 8
    attention_dropout: float = 0.
    feedforward_dropout: float = 0.
    feedforward_multiplier: int = 4
    norm_epsilon: float = 1e-6
    use_column_embedding: bool = True
    categorical_features_vocabulary: dict = None
    encode_categorical_values: bool = True
    numerical_normalization: str = "layer"



from tensorflow import keras
from teras.layers.node import ObliviousDecisionTree
from teras.utils.node import sparsemoid
from teras.activations import sparsemax
from teras.layers.node.node_feature_selector import NodeFeatureSelector
from teras.layerflow.models.node import NODE as _NodeLF
from teras.layers.common.head import ClassificationHead, RegressionHead
from teras.utils.types import UnitsValuesType


@keras.saving.register_keras_serializable("teras.models")
class NODE(_NodeLF):
    """
    Neural Oblivious Decision Tree (NODE) model
    based on the NODE architecture proposed by Sergei Popov et al.
    in paper Neural Oblivious Decision Ensembles for Deep Learning on Tabular Data

    Reference(s):
        https://arxiv.org/abs/1909.06312

    Args:
        input_dim: ``int``,
            Dimensionality of the input dataset,
            or the number of features in the dataset.

        num_layers: ``int``, default 8,
            Number of ObliviousDecisionTree layers to use in model

        num_trees: ``int``, default 128,
            Number of trees to use in each `ObliviousDecisionTree` layer

        depth: ``int``, default 6,
            Number of splits in every tree

        tree_dim: ``int``, default 1,
            Number of response channels in the response of individual tree

        max_features: ``int``,
            Maximum number of features to use. If None, all features in the input dataset will be used.

        input_dropout: ``float``, default 0.,
            Dropout rate to apply to inputs.

        choice_function:
            Function that computes feature weights s.t. f(tensor, dim).sum(dim) == 1
            By default, ``sparsemax`` is used.

        bin_function:
            Function that computes tree leaf weights.
            By default, ``sparsemoid`` is used.

        response_initializer: default "random_normal",
            Initializer for tree output tensor. Any format that is acceptable by the keras initializers.

        selection_logits_initializer: default "random_uniform",
            Initializer for logits that select features for the tree
            Both thresholds and scales are initialized with data-aware initialization function.

        threshold_init_beta: ``float``, default 1.0,
            Initializes threshold to a q-th quantile of data points
            where q ~ Beta(:threshold_init_beta:, :threshold_init_beta:)
            If this param is set to 1, initial thresholds will have the same distribution as data points
            If greater than 1 (e.g. 10), thresholds will be closer to median data value
            If less than 1 (e.g. 0.1), thresholds will approach min/max data values.

        threshold_init_cutoff: ``float``, default 1.0,
            Threshold log-temperatures initializer.
            By default(1.0), log-temperatures are initialized in such a way that all bin selectors
            end up in the linear region of sparse-sigmoid. The temperatures are then scaled by this parameter.
            Setting this value > 1.0 will result in some margin between data points and sparse-sigmoid cutoff value
            Setting this value < 1.0 will cause (1 - value) part of data points to end up in flat sparse-sigmoid region
            For instance, threshold_init_cutoff = 0.9 will set 10% points equal to 0.0 or 1.0
            Setting this value > 1.0 will result in a margin between data points and sparse-sigmoid cutoff value
            All points will be between (0.5 - 0.5 / threshold_init_cutoff) and (0.5 + 0.5 / threshold_init_cutoff)
    """
    def __init__(self,
                 input_dim: int,
                 num_layers: int = 8,
                 num_trees: int = 16,
                 depth: int = 6,
                 tree_dim: int = 1,
                 max_features: int = None,
                 input_dropout: float = 0.,
                 choice_function=None,
                 bin_function=None,
                 response_initializer="random_normal",
                 selection_logits_initializer="random_uniform",
                 threshold_init_beta: float = 1.0,
                 threshold_init_cutoff: float = 1.0,
                 **kwargs):
        choice_func = sparsemax if choice_function is None else choice_function
        bin_func = sparsemoid if bin_function is None else bin_function

        tree_layers = [ObliviousDecisionTree(num_trees=num_trees,
                                             depth=depth,
                                             tree_dim=tree_dim,
                                             choice_function=choice_func,
                                             bin_function=bin_func,
                                             response_initializer=response_initializer,
                                             selection_logits_initializer=selection_logits_initializer,
                                             threshold_init_beta=threshold_init_beta,
                                             threshold_init_cutoff=threshold_init_cutoff)
                       for _ in range(num_layers)]
        feature_selector = NodeFeatureSelector(data_dim=input_dim,
                                               max_features=max_features,
                                               name="feature_selector")
        dropout = keras.layers.Dropout(input_dropout,
                                       name="input_dropout")
        super().__init__(input_dim=input_dim,
                         tree_layers=tree_layers,
                         feature_selector=feature_selector,
                         dropout=dropout,
                         **kwargs)

        self.input_dim = input_dim
        self.num_layers = num_layers
        self.num_trees = num_trees
        self.depth = depth
        self.tree_dim = tree_dim
        self.max_features = max_features
        self.input_dropout = input_dropout
        self.choice_function = choice_function
        self.bin_function = bin_function
        self.choice_function = choice_function
        self.response_initializer = response_initializer
        self.selection_logits_initializer = selection_logits_initializer
        self.threshold_init_beta = threshold_init_beta
        self.threshold_init_cutoff = threshold_init_cutoff

    def get_config(self):
        config = {'name': self.name,
                  'trainable': self.trainable,
                  'input_dim': self.input_dim,
                  'num_layers': self.num_layers,
                  'num_trees': self.num_trees,
                  'depth': self.depth,
                  'tree_dim': self.tree_dim,
                  'max_features': self.max_features,
                  'input_dropout': self.input_dropout,
                  'choice_function': self.choice_function,
                  'bin_function': self.bin_function,
                  'response_initializer': self.response_initializer,
                  'selection_logits_initializer': self.selection_logits_initializer,
                  'threshold_init_beta': self.threshold_init_beta,
                  'threshold_init_cutoff': self.threshold_init_cutoff,
                  }
        return config

    @classmethod
    def from_config(cls, config):
        input_dim = config.pop("input_dim")
        return cls(input_dim=input_dim, **config)


@keras.saving.register_keras_serializable("teras.models")
class NODERegressor(NODE):
    """
    Neural Oblivious Decision Tree (NODE) Regressor model
    based on the NODE architecture proposed by Sergei Popov et al.
    in paper Neural Oblivious Decision Ensembles for Deep Learning on Tabular Data

    Reference(s):
        https://arxiv.org/abs/1909.06312

    Args:
        num_outputs: `int`, default 1,
            Number of regression outputs to predict.

        head_units_values: ``List[int]`` or ``Tuple[int]``, default None,
            Units values to use in the hidden layers in the Classification head.
            For each value in the list/tuple,
            a hidden layer of that dimensionality is added to the head.
            By default, no hidden layer is used.

        input_dim: ``int``,
            Dimensionality of the input dataset,
            or the number of features in the dataset.

        num_layers: ``int``, default 8,
            Number of ObliviousDecisionTree layers to use in model

        num_trees: ``int``, default 128,
            Number of trees to use in each `ObliviousDecisionTree` layer

        depth: ``int``, default 6,
            Number of splits in every tree

        tree_dim: ``int``, default 1,
            Number of response channels in the response of individual tree

        max_features: ``int``,
            Maximum number of features to use. If None, all features in the input dataset will be used.

        input_dropout: ``float``, default 0.,
            Dropout rate to apply to inputs.

        choice_function:
            Function that computes feature weights s.t. f(tensor, dim).sum(dim) == 1
            By default, ``sparsemax`` is used.

        bin_function:
            Function that computes tree leaf weights.
            By default, ``sparsemoid`` is used.

        response_initializer: default "random_normal",
            Initializer for tree output tensor. Any format that is acceptable by the keras initializers.

        selection_logits_initializer: default "random_uniform",
            Initializer for logits that select features for the tree
            Both thresholds and scales are initialized with data-aware initialization function.

        threshold_init_beta: ``float``, default 1.0,
            Initializes threshold to a q-th quantile of data points
            where q ~ Beta(:threshold_init_beta:, :threshold_init_beta:)
            If this param is set to 1, initial thresholds will have the same distribution as data points
            If greater than 1 (e.g. 10), thresholds will be closer to median data value
            If less than 1 (e.g. 0.1), thresholds will approach min/max data values.

        threshold_init_cutoff: ``float``, default 1.0,
            Threshold log-temperatures initializer.
            By default(1.0), log-temperatures are initialized in such a way that all bin selectors
            end up in the linear region of sparse-sigmoid. The temperatures are then scaled by this parameter.
            Setting this value > 1.0 will result in some margin between data points and sparse-sigmoid cutoff value
            Setting this value < 1.0 will cause (1 - value) part of data points to end up in flat sparse-sigmoid region
            For instance, threshold_init_cutoff = 0.9 will set 10% points equal to 0.0 or 1.0
            Setting this value > 1.0 will result in a margin between data points and sparse-sigmoid cutoff value
            All points will be between (0.5 - 0.5 / threshold_init_cutoff) and (0.5 + 0.5 / threshold_init_cutoff)
    """
    def __init__(self,
                 num_outputs: int = 1,
                 head_units_values: UnitsValuesType = None,
                 input_dim: int = None,
                 num_layers: int = 8,
                 num_trees: int = 16,
                 depth: int = 6,
                 tree_dim: int = 1,
                 max_features: int = None,
                 input_dropout: float = 0.,
                 choice_function=None,
                 bin_function=None,
                 response_initializer="random_normal",
                 selection_logits_initializer="random_uniform",
                 threshold_init_beta: float = 1.0,
                 threshold_init_cutoff: float = 1.0,
                 **kwargs):
        head = RegressionHead(num_outputs=num_outputs,
                              units_values=head_units_values,
                              normalization=None)
        super().__init__(input_dim=input_dim,
                         num_layers=num_layers,
                         num_trees=num_trees,
                         depth=depth,
                         tree_dim=tree_dim,
                         max_features=max_features,
                         input_dropout=input_dropout,
                         choice_function=choice_function,
                         bin_function=bin_function,
                         response_initializer=response_initializer,
                         selection_logits_initializer=selection_logits_initializer,
                         threshold_init_beta=threshold_init_beta,
                         threshold_init_cutoff=threshold_init_cutoff,
                         head=head,
                         **kwargs)
        self.num_outputs = num_outputs
        self.head_units_values = head_units_values

    def get_config(self):
        config = super().get_config()
        new_config = {'num_outputs': self.num_outputs,
                      'head_units_values': self.head_units_values,
                      }
        config.update(new_config)
        return config


@keras.saving.register_keras_serializable("teras.models")
class NODEClassifier(NODE):
    """
    Neural Oblivious Decision Tree (NODE) Classifier model
    based on the NODE architecture proposed by Sergei Popov et al.
    in paper Neural Oblivious Decision Ensembles for Deep Learning on Tabular Data

    Reference(s):
        https://arxiv.org/abs/1909.06312

    Args:
        num_classes: `int`, default 2,
            Number of classes to predict.

        head_units_values: ``List[int]`` or ``Tuple[int]``, default None,
            Units values to use in the hidden layers in the Classification head.
            For each value in the list/tuple,
            a hidden layer of that dimensionality is added to the head.
            By default, no hidden layer is used.

        input_dim: ``int``,
            Dimensionality of the input dataset,
            or the number of features in the dataset.

        num_layers: ``int``, default 8,
            Number of ObliviousDecisionTree layers to use in model

        num_trees: ``int``, default 128,
            Number of trees to use in each `ObliviousDecisionTree` layer

        depth: ``int``, default 6,
            Number of splits in every tree

        tree_dim: ``int``, default 1,
            Number of response channels in the response of individual tree

        max_features: ``int``,
            Maximum number of features to use. If None, all features in the input dataset will be used.

        input_dropout: ``float``, default 0.,
            Dropout rate to apply to inputs.

        choice_function:
            Function that computes feature weights s.t. f(tensor, dim).sum(dim) == 1
            By default, ``sparsemax`` is used.

        bin_function:
            Function that computes tree leaf weights.
            By default, ``sparsemoid`` is used.

        response_initializer: default "random_normal",
            Initializer for tree output tensor. Any format that is acceptable by the keras initializers.

        selection_logits_initializer: default "random_uniform",
            Initializer for logits that select features for the tree
            Both thresholds and scales are initialized with data-aware initialization function.

        threshold_init_beta: ``float``, default 1.0,
            Initializes threshold to a q-th quantile of data points
            where q ~ Beta(:threshold_init_beta:, :threshold_init_beta:)
            If this param is set to 1, initial thresholds will have the same distribution as data points
            If greater than 1 (e.g. 10), thresholds will be closer to median data value
            If less than 1 (e.g. 0.1), thresholds will approach min/max data values.

        threshold_init_cutoff: ``float``, default 1.0,
            Threshold log-temperatures initializer.
            By default(1.0), log-temperatures are initialized in such a way that all bin selectors
            end up in the linear region of sparse-sigmoid. The temperatures are then scaled by this parameter.
            Setting this value > 1.0 will result in some margin between data points and sparse-sigmoid cutoff value
            Setting this value < 1.0 will cause (1 - value) part of data points to end up in flat sparse-sigmoid region
            For instance, threshold_init_cutoff = 0.9 will set 10% points equal to 0.0 or 1.0
            Setting this value > 1.0 will result in a margin between data points and sparse-sigmoid cutoff value
            All points will be between (0.5 - 0.5 / threshold_init_cutoff) and (0.5 + 0.5 / threshold_init_cutoff)
    """
    def __init__(self,
                 num_classes: int = 2,
                 head_units_values: UnitsValuesType = None,
                 input_dim: int = None,
                 num_layers: int = 8,
                 num_trees: int = 16,
                 depth: int = 6,
                 tree_dim: int = 1,
                 max_features: int = None,
                 input_dropout: float = 0.,
                 choice_function=None,
                 bin_function=None,
                 response_initializer="random_normal",
                 selection_logits_initializer="random_uniform",
                 threshold_init_beta: float = 1.0,
                 threshold_init_cutoff: float = 1.0,
                 **kwargs):
        head = ClassificationHead(num_classes=num_classes,
                                  units_values=head_units_values,
                                  normalization=None)
        super().__init__(input_dim=input_dim,
                         num_layers=num_layers,
                         num_trees=num_trees,
                         depth=depth,
                         tree_dim=tree_dim,
                         max_features=max_features,
                         input_dropout=input_dropout,
                         choice_function=choice_function,
                         bin_function=bin_function,
                         response_initializer=response_initializer,
                         selection_logits_initializer=selection_logits_initializer,
                         threshold_init_beta=threshold_init_beta,
                         threshold_init_cutoff=threshold_init_cutoff,
                         head=head,
                         **kwargs)
        self.num_classes = num_classes
        self.head_units_values = head_units_values

    def get_config(self):
        config = super().get_config()
        new_config = {'num_classes': self.num_classes,
                      'head_units_values': self.head_units_values,
                      }
        config.update(new_config)
        return config



import tensorflow as tf
from tensorflow import keras
from teras.utils.dnfnet import (compute_total_number_of_literals,
                                compute_total_number_of_conjunctions,
                                compute_num_literals_per_formula,
                                create_conjunctions_indicator_matrix,
                                create_formulas_indicator_matrix,
                                dense_ndim_array_to_sparse_tensor,
                                tf_matmul_sparse_dense,
                                and_operator,
                                or_operator)
from teras.layers.dnfnet.dnfnet_localization import DNFNetLocalization
from teras.layers.dnfnet.dnfnet_feature_selection import DNFNetFeatureSelection
from teras.utils.types import (IntegerSequence,
                               FloatSequence)


@keras.saving.register_keras_serializable(package="teras.layers.dnfnet")
class DNNF(keras.layers.Layer):
    """
    Disjunctive Normal Neural Form (DNNF) layer
    is the  main building block of DNF-Net architecture.
    Based on the paper Net-DNF: Effective Deep Modeling of Tabular Data by Liran Katzir et al.

    Reference(s):
        https://openreview.net/forum?id=73WTGs96kho

    Args:
        num_formulas: ``int``, default 256,
            Number of DNF formulas. Each DNF formula is analogous to a tree in tree based ensembles.

        num_conjunctions_arr: ``List[int]`` or ``Tuple[int]``, default [6, 9, 12, 15],
            Conjunctions array.
            It is used in the computation of total number of literals as well as
            computation of number of literals per DNF formula.

        conjunctions_depth_arr: ``List[int]`` or ``Tuple[int]``, default [2, 4, 6],
            Conjunctions depth array.
            It is used in the computation of total number of literals as well as
            computation of number of literals per DNF formula.

        keep_feature_prob_arr: ``List[float]`` or ``Tuple[float]``, default [0.1, 0.3, 0.5, 0.7, 0.9],
            Feature probability array.
            It is used by the Feature Selection layer to randomly select a probability
            value that is used in the random selection of input features.

        elastic_net_beta: ``float``, default 0.4,
            Used in the computation of Elastic Net Regularization.

        binary_threshold_eps:   ``float``, default 1.0,
            Used in the computation of learnable mask.

        temperature: ``float``, default 2.0,
            Temperature value to use in the Localization layer.
            According to the paper, The inclusion of an adaptive temperature in this localization mechanism
            facilitates a data-dependent degree of exclusivity:
            at high temperatures, only a few DNNFs will handle an input instance whereas
            at low temperatures, more DNNFs will effectively participate in the ensemble.
    """

    def __init__(self,
                 num_formulas: int = 256,
                 num_conjunctions_arr: IntegerSequence = [6, 9, 12, 15],
                 conjunctions_depth_arr: IntegerSequence = [2, 4, 6],
                 keep_feature_prob_arr: FloatSequence = [0.1, 0.3, 0.5, 0.7, 0.9],
                 elastic_net_beta: float = 0.4,
                 binary_threshold_eps: float = 1.0,
                 temperature: float = 2.0,
                 **kwargs
                 ):
        super().__init__(**kwargs)
        self.num_formulas = num_formulas
        self.num_conjunctions_arr = num_conjunctions_arr
        self.conjunctions_depth_arr = conjunctions_depth_arr
        self.keep_feature_prob_arr = keep_feature_prob_arr
        self.elastic_net_beta = elastic_net_beta
        self.binary_threshold_eps = binary_threshold_eps
        self.temperature = temperature

        self.localization = DNFNetLocalization(num_formulas=self.num_formulas,
                                               temperature=self.temperature)

        self.loaded_input_masks = None
        self.feature_selection = None

    def build(self, input_shape):
        input_dim = input_shape[1]
        self.feature_selection = DNFNetFeatureSelection(input_dim=input_dim,
                                                        num_formulas=self.num_formulas,
                                                        keep_feature_prob_arr=self.keep_feature_prob_arr,
                                                        elastic_net_beta=self.elastic_net_beta,
                                                        binary_threshold_eps=self.binary_threshold_eps)

        self.total_number_of_literals = compute_total_number_of_literals(self.num_formulas,
                                                                         self.num_conjunctions_arr,
                                                                         self.conjunctions_depth_arr)
        num_literals_per_formula_arr = compute_num_literals_per_formula(self.num_conjunctions_arr,
                                                                        self.conjunctions_depth_arr)
        total_number_of_conjunctions = compute_total_number_of_conjunctions(self.num_formulas,
                                                                            self.num_conjunctions_arr)

        self.learnable_binary_mask, self.literals_random_mask, self.elastic_net_reg = self.feature_selection(
            num_literals_per_formula_arr)

        self.bias = tf.Variable(name='literals_bias',
                                shape=tf.shape(self.literals_random_mask)[1],
                                initial_value=tf.fill(dims=(tf.shape(self.literals_random_mask)[1],), value=0.))
        self.weight = self.add_weight(name='literals_weights',
                                      shape=tf.shape(self.literals_random_mask),
                                      initializer=keras.initializers.glorot_normal())
        self.weight = tf.multiply(self.weight, self.literals_random_mask)

        # 'b' is just a prop to hold intermediate values since we can't assign values to a tensor,
        # so we define a variable here with trainable argument set to False.
        # same is the case with 'c' below.
        b = tf.Variable(tf.cast(tf.zeros(tf.shape(self.literals_random_mask)[1]),
                                dtype=tf.bool),
                        trainable=False)
        conjunctions_indicator_matrix = create_conjunctions_indicator_matrix(b,
                                                                             total_number_of_conjunctions,
                                                                             self.conjunctions_depth_arr,
                                                                             )
        conjunctions_indicator_matrix = tf.cast(conjunctions_indicator_matrix, dtype=tf.float32)
        self.conjunctions_indicator_sparse_matrix = dense_ndim_array_to_sparse_tensor(conjunctions_indicator_matrix)
        self.and_bias = tf.reduce_sum(conjunctions_indicator_matrix,
                                      axis=0)

        c = tf.Variable(tf.cast(tf.zeros(total_number_of_conjunctions),
                                dtype=tf.bool))
        formulas_indicator_matrix = create_formulas_indicator_matrix(c,
                                                                     self.num_formulas,
                                                                     self.num_conjunctions_arr)
        formulas_indicator_matrix = tf.cast(formulas_indicator_matrix, dtype=tf.float32)
        self.formulas_indicator_sparse_matrix = dense_ndim_array_to_sparse_tensor(formulas_indicator_matrix)
        self.or_bias = tf.reduce_sum(formulas_indicator_matrix,
                                     axis=0)

    def call(self, inputs):
        x = inputs
        # Creating affine literals
        out_literals = tf.tanh(tf.add(tf.matmul(x, tf.multiply(self.weight, self.learnable_binary_mask)), self.bias))

        # Soft Conjunctions
        out_conjunctions = and_operator(tf_matmul_sparse_dense(self.conjunctions_indicator_sparse_matrix, out_literals),
                                        d=self.and_bias)
        out_DNNFs = or_operator(tf_matmul_sparse_dense(self.formulas_indicator_sparse_matrix, out_conjunctions),
                                d=self.or_bias)
        out_DNNFs = tf.reshape(out_DNNFs, shape=(-1, self.num_formulas))

        # Localization
        loc = self.localization(x)
        out_DNNFs = tf.multiply(out_DNNFs, loc)
        return out_DNNFs

    def get_config(self):
        config = super().get_config()
        config.update({'num_formulas': self.num_formulas,
                       'num_conjunctions_arr': self.num_conjunctions_arr,
                       'conjunctions_depth_arr': self.conjunctions_depth_arr,
                       'keep_feature_prob_arr': self.keep_feature_prob_arr,
                       'elastic_net_beta': self.elastic_net_beta,
                       'binary_threshold_eps': self.binary_threshold_eps,
                       'temperature': self.temperature}
                      )
        return config



import tensorflow as tf
from tensorflow import keras
from teras.utils import get_normalization_layer
from teras.utils.types import NormalizationType


@keras.saving.register_keras_serializable(package="teras.layers")
class NumericalFeatureNormalization(keras.layers.Layer):
    """
    NumericalFeatureNormalization layer that applies specified
    type of normalization over the numerical features only.


    Args:
        features_metadata: ``dict``,
            a nested dictionary of metadata for features where
            categorical sub-dictionary is a mapping of categorical feature names to a tuple of
            feature indices and the lists of unique values (vocabulary) in them,
            while numerical dictionary is a mapping of numerical feature names to their indices.
            `{feature_name: (feature_idx, vocabulary)}` for feature in categorical features.
            `{feature_name: feature_idx}` for feature in numerical features.
            You can get this metadata dictionary by calling
                >>> from teras.utils import get_features_metadata_for_embedding
                >>> features_metadata = get_features_metadata_for_embedding(dataset,
                ..                                                            categorical_features,
                ..                                                            numerical_features)
        normalization: ``str``, default "layer",
            Normalization type to apply over the numerical features.
            By default, ``LayerNormalization`` is applied.
            Allowed values ["layer", "batch", "unit", "group"] or their full names,
            e.g. "LayerNormalization".
            Note: The names are case-insensitive.
    """
    def __init__(self,
                 features_metadata: dict,
                 normalization: NormalizationType = "layer",
                 **kwargs):
        super().__init__(**kwargs)
        if not len(features_metadata["numerical"]) > 0:
            raise ValueError("`features_metadata` contains no numerical features. "
                             "Either you forgot to pass numerical features names list to the "
                             "`get_features_metadata_for_embedding` or the dataset does not contain "
                             "any numerical features to begin with. \n"
                             "In either case, "
                             "`NumericalFeatureNormalization` cannot be called on inputs with no numerical features. ")
        self.features_metadata = features_metadata
        self.normalization = normalization
        self.norm = get_normalization_layer(normalization)

        self._numerical_features_idx = list(self.features_metadata["numerical"].values())

    def call(self, inputs):
        numerical_features = tf.gather(inputs,
                                       indices=self._numerical_features_idx,
                                       axis=1)
        numerical_features = self.norm(numerical_features)
        return numerical_features

    def get_config(self):
        config = super().get_config()
        if isinstance(self.normalization, str):
            normalization_serialized = self.normalization
        else:
            normalization_serialized = keras.layers.serialize(self.normalization)

        config.update({'features_metadata': self.features_metadata,
                       'normalization': normalization_serialized,
                       })
        return config



import tensorflow as tf
from tensorflow import keras
from tensorflow.keras import losses


def info_nce_loss(real_projection_outputs=None,
                  augmented_projection_outputs=None,
                  temperature: float = 0.7,
                  lambda_: float = 0.5):
    """
    Contrastive Info NCE loss.
    In SAINT architecture, it is used in combination with the `denoising` loss.
    Args:
        projection_outputs_original: Outputs of projection head over encodings of original inputs
        projection_outputs_augmented: Outputs of projection head over the encodings of augmented inputs
        temperature: `float`, default 0.7,
            Temperature value used in the computation of the InfoNCE contrastive loss.
        lambda_: `float`, default 0.5,
           Used in combining the two losses.

    Returns:
        Info NCE loss.
    """
    batch_size = tf.shape(real_projection_outputs)[0]
    labels = tf.range(batch_size)
    logits_ab = tf.matmul(real_projection_outputs, augmented_projection_outputs, transpose_b=True) / temperature
    logits_ba = tf.matmul(augmented_projection_outputs, real_projection_outputs, transpose_b=True) / temperature
    loss_a = losses.SparseCategoricalCrossentropy(from_logits=True)(y_true=labels,
                                                                    y_pred=logits_ab)
    loss_b = losses.SparseCategoricalCrossentropy(from_logits=True)(y_true=labels,
                                                                    y_pred=logits_ba)
    loss = lambda_ * (loss_a + loss_b) / 2
    return loss


def denoising_loss(real_samples=None,
                   reconstructed_samples=None,
                   categorical_features_metadata: dict = None):
    """
    Since we apply categorical and numerical embedding layers
    separately and then combine them into a new features matrix
    this effectively makes the first k features in the outputs
    categorical (since categorical embeddings are applied first)
    and all other features numerical.
    Here, k = num_categorical_features

    Args:
        real_samples: `tf.Tensor`,
            Samples drawn from the original dataset.
        reconstructed_samples: `tf.Tensor`,
            Samples reconstructed by the reconstruction head.
        num_categorical_features: `int`,
            Number of categorical features in the dataset.
            If there are no categorical features, specify 0.
    """
    if real_samples is None:
        raise ValueError("`real_samples` cannot be None. "
                         "You must pass the samples drawn from the original dataset. ")

    if reconstructed_samples is None:
        raise ValueError("`reconstructed_samples` cannot be None. "
                         "You must pass the samples reconstructed by the ReconstructionHead.")

    if categorical_features_metadata is None:
        raise ValueError("`categorical_features_metadata` cannot be None. ")

    num_features = tf.shape(real_samples)[1]
    num_categorical_features = len(categorical_features_metadata)
    num_categories_per_feature = list(map(lambda x: len(x[1]), categorical_features_metadata.values()))
    total_categories = sum(num_categories_per_feature)
    loss = 0.
    # The reconstructed samples have dimensions equal to number of categorical features + number of categories in all
    # the categorical features combined.
    # In other words, each categorical feature gets expanded into `number of categories` features due to softmax layer
    # during the reconstruction phase.
    # Since the order they occur in is still the same
    # except for that the first `total_categories` number of features are categorical and following are numerical
    for current_idx, (feature_idx, _) in enumerate(categorical_features_metadata.values()):
        num_categories = num_categories_per_feature[current_idx]
        loss += tf.keras.losses.SparseCategoricalCrossentropy(from_logits=True)(
                                    y_true=real_samples[:, feature_idx],  # real_samples have the original feature order
                                    y_pred=reconstructed_samples[:, current_idx: current_idx + num_categories])
        current_idx += 1
    # Check if there are numerical features -- if so, compute the mse loss
    mse_loss = tf.cond(pred=num_categorical_features < num_features,
                       true_fn=lambda: tf.reduce_sum(tf.losses.mse(real_samples[:, num_categorical_features:],
                                                                   reconstructed_samples[:, total_categories:])),
                       false_fn=lambda: 0.
                       )
    loss += mse_loss
    return loss



import tensorflow as tf
from tensorflow import keras
from teras.utils.types import Number


@keras.saving.register_keras_serializable(package="teras.layers.preprocessing")
class LabelEncoding(keras.layers.Layer):
    """
    Standalone Encoding layer
    Essentially it acts as a Label encoder.
    It maps the numeric values (int/float) to their indices.

    Args:
        inputs: ``numpy ndarray`` or ``tf.Tensor``,
            Inputs with categorical features that should be encoded

        concatenate_numerical_features: ``bool``, default False,
            Whether to concatenate numerical features
            to the encoded categorical features.

        keep_features_order: ``bool``, default True,
            If True, the returning dataset will have the features in the same
            order as the input.
            If False, the returning dataset, will have k categorical features
            first and then the numerical features follow in the order they appear
            in the dataset i.e. first come, first served essentially.

            This parameter will only be taken into account when the
            `concatenate_numerical_features` is set to True,
            as otherwise returning dataset will only contain categorical features
            in order that they appear in the inputs.
    """
    def __init__(self,
                 features_metadata: dict,
                 concatenate_numerical_features: bool = False,
                 keep_features_order: bool = True,
                 **kwargs):
        super().__init__(**kwargs)
        self.features_metadata = features_metadata
        self.features_metadata["categorical"] = self.features_metadata["categorical"]
        self.concatenate_numerical_features = concatenate_numerical_features
        self.keep_features_order = keep_features_order

        self._categorical_features_idx = list(map(lambda x: x[0], self.features_metadata["categorical"].values()))
        self._num_categorical_features = len(self._categorical_features_idx)
        self._numerical_features_idx = list(self.features_metadata["numerical"].values())
        self._num_numerical_features = len(self._numerical_features_idx)
        self._num_features = self._num_numerical_features + self._num_categorical_features

        self.lookup_tables = self._get_lookup_tables()

    def _get_lookup_tables(self):
        """Lookup tables for each categorical feature"""
        lookup_tables = []
        for feature_name, (feature_idx, vocabulary) in self.features_metadata["categorical"].items():
            # Lookup Table to convert int/float values to integer indices
            if isinstance(vocabulary[0], Number):
                lookup = keras.layers.IntegerLookup(vocabulary=vocabulary,
                                                    mask_token=None,
                                                    output_mode="int",
                                                    )
            else:
                raise TypeError("`LabelEncoding` layer can only encode values of type int "
                                f"but received type: {type(vocabulary[0])} for feature {feature_name}.")
            lookup_tables.append(lookup)
        return lookup_tables

    def call(self, inputs):
        if self.concatenate_numerical_features:
            encoded_features = tf.TensorArray(size=self._num_features,
                                              dtype=tf.float32)
        else:
            encoded_features = tf.TensorArray(size=self._num_categorical_features,
                                              dtype=tf.float32)
        current_idx = 0
        for feature_idx in self._categorical_features_idx:
            feature = tf.expand_dims(inputs[:, feature_idx], 1)
            # Convert string input values to integer indices
            lookup = self.lookup_tables[current_idx]
            feature = tf.cast(lookup(feature), dtype=tf.float32)
            encoded_features = encoded_features.write(index=feature_idx if (self.concatenate_numerical_features
                                                                            and self.keep_features_order) else current_idx,
                                                      value=feature)
            current_idx += 1

        if self.concatenate_numerical_features:
            for i, feature_idx in enumerate(self._numerical_features_idx):
                feature = tf.cast(tf.expand_dims(inputs[:, feature_idx], 1), tf.float32)
                encoded_features = encoded_features.write(index=feature_idx if self.keep_features_order else current_idx,
                                                          value=feature)
                current_idx += 1

        encoded_features = tf.squeeze(encoded_features.stack())
        encoded_features = tf.transpose(encoded_features)

        if self.concatenate_numerical_features:
            encoded_features.set_shape((None, self._num_features))
        else:
            encoded_features.set_shape((None, self._num_categorical_features))
        return encoded_features

    def get_config(self):
        config = super().get_config()
        new_config = {'features_metadata': self.features_metadata,
                      'concatenate_numerical_features': self.concatenate_numerical_features,
                      'keep_features_order': self.keep_features_order}
        config.update(new_config)
        return config



from tensorflow import keras
from teras.utils.types import ActivationType, InitializerType


@keras.saving.register_keras_serializable(package="teras.layers.gain")
class GAINGeneratorBlock(keras.layers.Layer):
    """
    ``GAINGeneratorBlock`` that is used in construction of hidden block
    inside the ``GAINGenerator`` model of ``GAIN`` architecture.

    Args:
        units: ``int``,
            Hidden dimensionality of the generator block

        activation: ``str``, ``keras.initializer.Initializer``, default "relu",
            Activation function to use in the generator block

        kernel_initializer: ``str``, ``keras.initializer.Initializer``, default "glorot_normal",
            Kernel initializer for the generator block
    """
    def __init__(self,
                 units: int,
                 activation: ActivationType = "relu",
                 kernel_initializer: InitializerType = "glorot_normal",
                 **kwargs):
        super().__init__(**kwargs)
        self.units = units
        self.activation = activation
        self.kernel_initializer = kernel_initializer
        self.dense = keras.layers.Dense(units,
                                        activation=self.activation,
                                        kernel_initializer=self.kernel_initializer)

    def call(self, inputs):
        return self.dense(inputs)

    def get_config(self):
        config = super().get_config()

        activation_serialized = self.activation
        if not isinstance(self.activation, str):
            activation_serialized = keras.layers.deserialize(self.activation)

        kernel_initializer_serialized = self.kernel_initializer
        if not isinstance(self.activation, str):
            kernel_initializer_serialized = keras.layers.deserialize(self.kernel_initializer)

        config.update({'units': self.units,
                       'activation': activation_serialized,
                       'kernel_initializer': kernel_initializer_serialized,
                       }
                      )
        return config

    @classmethod
    def from_config(cls, config):
        units = config.pop("units")
        return cls(units, **config)



from tensorflow import keras
from teras.layerflow.layers.saint.saint_encoder import SAINTEncoder as _SAINTEncoderLF
from teras.layers.saint.saint_transformer import SAINTTransformer


@keras.saving.register_keras_serializable(package="teras.layers.saint")
class SAINTEncoder(_SAINTEncoderLF):
    """
    SAINTEncoder for the SAINT architecture as proposed by
    Gowthami Somepalli et al. in the paper
    SAINT: Improved Neural Networks for Tabular Data
    via Row Attention and Contrastive Pre-Training.
    It simply stacks N transformer layers and applies them to the outputs
    of the embedded features.

    It differs from the typical Encoder block only in that the Transformer
    layer is a bit different from the regular Transformer layer used in the
    Transformer based architectures as it uses multi-head inter-sample attention,
    in addition to the regular mutli-head attention for features.

    Reference(s):
        https://arxiv.org/abs/2106.01342

    Args:
        data_dim: ``int``,
            Dimensionality of the input dataset,
            or the total number of features in the dataset.

        num_transformer_layer: ``int``, default 6,
            Number of transformer layers to use in the Encoder

        embedding_dim: ``int``, default 32,
            Embedding dimensions used to embedd numerical and
            categorical features. These server as the key dimensions
            in the MultiHeadAttention layer.

        num_attention_heads: ``int``, default 8,
            Number of heads to use in the typical ``MultiHeadAttention``
            layer that will be applied over features.

        num_inter_sample_attention_heads: ``int``, default 8,
            Number of heads to use in the ``MultiHeadInterSampleAttention``
            that will be applied over rows

        embedding_dim: ``int``, default 32,
            Embedding dimensions. These will also serve as key dimensions
             for the attention layers

        attention_dropout: ``float``, default 0.1,
            Dropout rate for ``MultiHeadAttention`` which is applied over
             features.

        inter_sample_attention_dropout: ``float``, default 0.1,
            Dropout rate for ``MultiHeadInterSampleAttention`` which is
            applied over rows.

        feedforward_dropout: ``float``, default 0.1,
            Dropout rate for the ``Dropout`` layer that is part of the
            ``FeedForward`` layer.

        feedforward_multiplier: ``int``, default 4.
            Multiplier that is multiplied with the ``embedding_dim``
            and the resultant value is used as hidden dimensions value for the
            hidden layer in the ``FeedForward`` layer.

        apply_attention_to_features: ``bool``, default True,
            Whether to apply attention over features.
            If True, the regular ``MultiHeadAttention`` layer will be applied
            over features.

        apply_attention_to_rows: ``bool``, default True,
            Whether to apply attention over rows.
            If True, the ``MultiHeadInterSampleAttention`` will apply attention
            over rows.
            NOTE: It is strongly recommended to keep both as True, but you
            can turn one off for experiment's sake.
            Also, note that, both CANNOT be False at the same time!
    """
    def __init__(self,
                 data_dim: int,
                 num_transformer_layers: int = 6,
                 embedding_dim: int = 32,
                 num_attention_heads: int = 8,
                 num_inter_sample_attention_heads: int = 8,
                 attention_dropout: float = 0.1,
                 inter_sample_attention_dropout: float = 0.1,
                 feedforward_dropout: float = 0.1,
                 feedforward_multiplier: int = 4,
                 norm_epsilon: float = 1e-6,
                 apply_attention_to_features: bool = True,
                 apply_attention_to_rows: bool = True,
                 **kwargs):
        if not apply_attention_to_features and not apply_attention_to_rows:
            raise ValueError("`apply_attention_to_features` and `apply_attention_to_rows` both cannot be False "
                             "at the same time. You must set at least one to True if not both. "
                             f"Received: `apply_attention_to_features`={apply_attention_to_features}, "
                             f"`apply_attention_to_rows`={apply_attention_to_rows}")

        saint_transformer_layers = keras.models.Sequential(name="saint_transformer_layers")
        for i in range(num_transformer_layers):
            saint_transformer_layers.add(SAINTTransformer(
                data_dim=data_dim,
                embedding_dim=embedding_dim,
                num_attention_heads=num_attention_heads,
                num_inter_sample_attention_heads=num_inter_sample_attention_heads,
                attention_dropout=attention_dropout,
                inter_sample_attention_dropout=inter_sample_attention_dropout,
                feedforward_dropout=feedforward_dropout,
                feedforward_multiplier=feedforward_multiplier,
                apply_attention_to_features=apply_attention_to_features,
                apply_attention_to_rows=apply_attention_to_rows,
                name=f"saint_transformer_layer_{i}"))

        super().__init__(saint_transformer_layers=saint_transformer_layers,
                         **kwargs)

        self.data_dim = data_dim
        self.num_transformer_layers = num_transformer_layers
        self.embedding_dim = embedding_dim
        self.num_attention_heads = num_attention_heads
        self.num_inter_sample_attention_heads = num_inter_sample_attention_heads
        self.attention_dropout = attention_dropout
        self.inter_sample_attention_dropout = inter_sample_attention_dropout
        self.feedforward_dropout = feedforward_dropout
        self.feedforward_multiplier = feedforward_multiplier
        self.norm_epsilon = norm_epsilon
        self.apply_attention_to_features = apply_attention_to_features
        self.apply_attention_to_rows = apply_attention_to_rows

    def get_config(self):
        config = {'name': self.name,
                  'trainable': self.trainable,
                  'data_dim': self.data_dim,
                  'num_transformer_layers': self.num_transformer_layers,
                  'embedding_dim': self.embedding_dim,
                  'num_attention_heads': self.num_attention_heads,
                  'num_inter_sample_attention_heads': self.num_inter_sample_attention_heads,
                  'attention_dropout': self.attention_dropout,
                  'inter_sample_attention_dropout': self.inter_sample_attention_dropout,
                  'feedforward_dropout': self.feedforward_dropout,
                  'feedforward_multiplier': self.feedforward_multiplier,
                  'norm_epsilon': self.norm_epsilon,
                  'apply_attention_to_features': self.apply_attention_to_features,
                  'apply_attention_to_rows': self.apply_attention_to_rows,
                  }
        return config

    @classmethod
    def from_config(cls, config):
        # data_dim is the only positional argument
        data_dim = config.pop("data_dim")
        return cls(data_dim=data_dim, **config)



import tensorflow as tf
from tensorflow import keras


@keras.saving.register_keras_serializable(package="teras.layers")
class NumericalFeaturesExtractor(keras.layers.Layer):
    """
    NumericalFeaturesExtractor layer extracts numerical features as is.
    It helps us build functional model, in case we don't want to apply
    any special layer to numerical features but still want to extract
    numerical features to concatenate them with the categorical features.

    Args:
        features_metadata: ``dict``,
            A nested dictionary of metadata for features where
            categorical sub-dictionary is a mapping of categorical feature names to a tuple of
            feature indices and the lists of unique values (vocabulary) in them,
            while numerical dictionary is a mapping of numerical feature names to their indices.
            ``{feature_name: (feature_idx, vocabulary)}`` for feature in categorical features.
            ``{feature_name: feature_idx}`` for feature in numerical features.
            You can get this dictionary from
                >>> from teras.utils import get_features_metadata_for_embedding
                >>> metadata_dict = get_features_metadata_for_embedding(dataframe,
                ..                                                      categorical_features,
                ..                                                      numerical_features)

    Raises:
        If no numerical features exist, this layer raises a ``ValueError``.
    """
    def __init__(self,
                 features_metadata: dict,
                 **kwargs):
        super().__init__(**kwargs)
        self.features_metadata = features_metadata
        if not len(self.features_metadata["numerical"]) > 0:
            raise ValueError("`features_metadata` contains no numerical features. "
                             "Either you forgot to pass numerical features names list to the "
                             "`get_features_metadata_for_embedding` or the dataset does not contain "
                             "any numerical features to begin with. \n"
                             "In either case, "
                             "`NumericalFeaturesExtractor` cannot be called on inputs with no numerical features. ")
        self._numerical_features_idx = list(self.features_metadata["numerical"].values())

    def call(self, inputs):
        numerical_features = tf.gather(inputs,
                                       indices=self._numerical_features_idx,
                                       axis=1)
        return numerical_features



from tensorflow import keras
from teras.config.tabtransformer import TabTransformerConfig
from teras.layerflow.models.tabtransformer import TabTransformer as TabTransformerLF
from teras.layers.common.transformer import Encoder
from teras.layers.categorical_feature_embedding import CategoricalFeatureEmbedding
from teras.layers.numerical_feature_normalization import NumericalFeatureNormalization
from teras.layers.tabtransformer.tabtransformer_column_embedding import TabTransformerColumnEmbedding
from teras.layers.common.head import ClassificationHead, RegressionHead
from teras.utils.types import UnitsValuesType, NormalizationType


@keras.saving.register_keras_serializable("keras.models")
class TabTransformer(TabTransformerLF):
    """
    TabTransformer architecture as proposed by Xin Huang et al.
    in the paper,
    TabTransformer: Tabular Data Modeling Using Contextual Embeddings.

    TabTransformer, a novel deep tabular data modeling architecture for
    supervised and semi-supervised learning.
    The TabTransformer is built upon self-attention based Transformers.
    The Transformer layers transform the embeddings of categorical features
    into robust contextual embeddings to achieve higher prediction accuracy.

    Reference(s):
        https://arxiv.org/abs/2012.06678

    Args:
        input_dim: ``int``,
            Dimensionality of the input dataset,
            or the number of features in the dataset.

        features_metadata: ``dict``,
            A nested dictionary of metadata for features where
            categorical sub-dictionary is a mapping of categorical feature names to a tuple of
            feature indices and the lists of unique values (vocabulary) in them,
            while numerical dictionary is a mapping of numerical feature names to their indices.
            `{feature_name: (feature_idx, vocabulary)}` for feature in categorical features.
            `{feature_name: feature_idx}` for feature in numerical features.
            You can get this dictionary from
                >>> from teras.utils import get_features_metadata_for_embedding
                >>> metadata_dict = get_features_metadata_for_embedding(dataframe,
                ..                                                      categorical_features,
                ..                                                      numerical_features)

        embedding_dim: ``int``, default 32,
            Dimensionality of the learnable feature embeddings for categorical features.

        num_transformer_layers: ``int``, default 6,
            Number of transformer layers to use in the ``Encoder`` layer.
            The ``Encoder`` is used to contextualize the learned feature embeddings.

        num_attention_heads: ``int``, default 8,
            Number of attention heads to use in the ``MultiHeadSelfAttention`` layer
            that is part of the ``Transformer`` layer which in turn is part of the
            ``Encoder`` layer.

        attention_dropout: ``float``, default 0.0,
            Dropout rate to use in the ``MultiHeadSelfAttention`` layer in the transformer layer.

        feedforward_dropout: ``float``, default 0.0,
            Dropout rate to use for the ``Dropout`` layer in the ``FeedForward`` layer.

        feedforward_multiplier: ``int``, default 4.
            Multiplier that is multiplied with the ``embedding_dim``
            and the resultant value is used as hidden dimensions value for the
            hidden layer in the ``FeedForward`` layer.

        norm_epsilon: ``float``, default 1e-6,
            A very small number used for normalization in the ``LayerNormalization`` layer.

        use_column_embedding: ``bool``, default True,
            Whether to use the novel ``ColumnEmbedding`` layer proposed in the
            ``TabTransformer`` architecture for the categorical features.
            The ``ColumnEmbedding`` layer is an alternative to positional encoding that is applied
            in the Transformers in Natural Langauge Processing application settings.

        encode_categorical_values: ``bool``, default True,
            Whether to (label) encode categorical values,
            If you've already encoded the categorical values using for instance
            Label/Ordinal encoding, you should set this to False,
            otherwise leave it as True.
            In the case of True, categorical values will be mapped to integer indices
            using keras's ``IntegerLookup`` layer.

        numerical_normalization: ``str`` or ``keras.layes.Layer``, default "layer",
            Normalization to use for the numerical features.
"""
    def __init__(self,
                 input_dim: int,
                 features_metadata: dict,
                 embedding_dim: int = TabTransformerConfig.embedding_dim,
                 num_transformer_layers: int = TabTransformerConfig.num_transformer_layers,
                 num_attention_heads: int = TabTransformerConfig.num_attention_heads,
                 attention_dropout: float = TabTransformerConfig.attention_dropout,
                 feedforward_dropout: float = TabTransformerConfig.feedforward_dropout,
                 feedforward_multiplier: int = TabTransformerConfig.feedforward_multiplier,
                 norm_epsilon: float = TabTransformerConfig.norm_epsilon,
                 use_column_embedding: bool = TabTransformerConfig.use_column_embedding,
                 encode_categorical_values: bool = TabTransformerConfig.encode_categorical_values,
                 numerical_normalization: NormalizationType = TabTransformerConfig.numerical_normalization,
                 **kwargs
                 ):

        if features_metadata is None:
            raise ValueError(f"""
            `features_metadata` is required for embedding features and hence cannot be None.            
            You can get this features_metadata dictionary by calling
            `teras.utils.get_categorical_features_vocabulary(dataset, categorical_features, numerical_features)`
            Received, `features_metadata`: {features_metadata}
            """)

        categorical_feature_embedding = CategoricalFeatureEmbedding(
            features_metadata=features_metadata,
            embedding_dim=embedding_dim,
            encode=encode_categorical_values
        )
        column_embedding = TabTransformerColumnEmbedding(embedding_dim=embedding_dim,
                                                         num_categorical_features=len(features_metadata["categorical"]))
        encoder = Encoder(num_transformer_layers=num_transformer_layers,
                          num_attention_heads=num_attention_heads,
                          embedding_dim=embedding_dim,
                          attention_dropout=attention_dropout,
                          feedforward_dropout=feedforward_dropout,
                          feedforward_multiplier=feedforward_multiplier,
                          norm_epsilon=norm_epsilon)
        numerical_feature_normalization = NumericalFeatureNormalization(features_metadata=features_metadata,
                                                                        normalization=numerical_normalization)

        super().__init__(input_dim=input_dim,
                         categorical_feature_embedding=categorical_feature_embedding,
                         column_embedding=column_embedding,
                         encoder=encoder,
                         numerical_feature_normalization=numerical_feature_normalization,
                         **kwargs)

        self.input_dim = input_dim
        self.features_metadata = features_metadata
        self.embedding_dim = embedding_dim
        self.num_transformer_layers = num_transformer_layers
        self.num_attention_heads = num_attention_heads
        self.attention_dropout = attention_dropout
        self.feedforward_dropout = feedforward_dropout
        self.feedforward_multiplier = feedforward_multiplier
        self.use_column_embedding = use_column_embedding
        self.norm_epsilon = norm_epsilon
        self.encode_categorical_values = encode_categorical_values
        self.numerical_normalization = numerical_normalization

    def get_config(self):
        numerical_normalization_serialized = self.numerical_normalization
        if not isinstance(numerical_normalization_serialized, str):
            numerical_normalization_serialized = keras.layers.serialize(numerical_normalization_serialized)
        config = {'name': self.name,
                  'trainable': self.trainable,
                  'input_dim': self.input_dim,
                  'features_metadata': self.features_metadata,
                  'embedding_dim': self.embedding_dim,
                  'num_transformer_layers': self.num_transformer_layers,
                  'num_attention_heads': self.num_attention_heads,
                  'attention_dropout': self.attention_dropout,
                  'feedforward_dropout': self.feedforward_dropout,
                  'feedforward_multiplier': self.feedforward_multiplier,
                  'norm_epsilon': self.norm_epsilon,
                  'use_column_embedding': self.use_column_embedding,
                  'encode_categorical_values': self.encode_categorical_values,
                  'numerical_normalization': numerical_normalization_serialized,
                  }
        return config

    @classmethod
    def from_config(cls, config):
        # Super class i.e. TabTransformerLF's ``from_config`` method  would try
        # to build the TabTransformerLF class expecting configuration to contain
        # sublayers' configs such as ``CategoricalFeatureEmebedding`` layer's config etc.
        # Hence, overriding the from_config method here is imperative.
        #
        # !NOTE: Since in this default/parametric API version, we don't take any
        # sublayers as arguments, so, we don't need to deserialize anything.
        input_dim = config.pop("input_dim")
        features_metadata = config.pop("features_metadata")
        return cls(input_dim=input_dim,
                   features_metadata=features_metadata,
                   **config)


class TabTransformerClassifier(TabTransformer):
    """
    TabTransformerClassifier based on the TabTransformer architecture
    as proposed by Xin Huang et al. in the paper,
    TabTransformer: Tabular Data Modeling Using Contextual Embeddings.

    TabTransformer, a novel deep tabular data modeling architecture for
    supervised and semi-supervised learning.
    The TabTransformer is built upon self-attention based Transformers.
    The Transformer layers transform the embeddings of categorical features
    into robust contextual embeddings to achieve higher prediction accuracy.

    Reference(s):
        https://arxiv.org/abs/2012.06678

    Args:
        num_classes: ``int``, default 2,
            Number of classes to predict.

        head_units_values: ``List[int]`` or ``Tuple[int]``, default [64, 32],
            Units values to use in the hidden layers in the Classification head.
            For each value in the list/tuple,
            a hidden layer of that dimensionality is added to the head.

        input_dim: ``int``,
            Dimensionality of the input dataset,
            or the number of features in the dataset.

        features_metadata: ``dict``,
            A nested dictionary of metadata for features where
            categorical sub-dictionary is a mapping of categorical feature names to a tuple of
            feature indices and the lists of unique values (vocabulary) in them,
            while numerical dictionary is a mapping of numerical feature names to their indices.
            `{feature_name: (feature_idx, vocabulary)}` for feature in categorical features.
            `{feature_name: feature_idx}` for feature in numerical features.
            You can get this dictionary from
                >>> from teras.utils import get_features_metadata_for_embedding
                >>> metadata_dict = get_features_metadata_for_embedding(dataframe,
                ..                                                      categorical_features,
                ..                                                      numerical_features)

        embedding_dim: ``int``, default 32,
            Dimensionality of the learnable feature embeddings for categorical features.

        num_transformer_layers: ``int``, default 6,
            Number of transformer layers to use in the ``Encoder`` layer.
            The ``Encoder`` is used to contextualize the learned feature embeddings.

        num_attention_heads: ``int``, default 8,
            Number of attention heads to use in the ``MultiHeadSelfAttention`` layer
            that is part of the ``Transformer`` layer which in turn is part of the
            ``Encoder`` layer.

        attention_dropout: ``float``, default 0.0,
            Dropout rate to use in the ``MultiHeadSelfAttention`` layer in the transformer layer.

        feedforward_dropout: ``float``, default 0.0,
            Dropout rate to use for the ``Dropout`` layer in the ``FeedForward`` layer.

        feedforward_multiplier: ``int``, default 4.
            Multiplier that is multiplied with the ``embedding_dim``
            and the resultant value is used as hidden dimensions value for the
            hidden layer in the ``FeedForward`` layer.

        norm_epsilon: ``float``, default 1e-6,
            A very small number used for normalization in the ``LayerNormalization`` layer.

        use_column_embedding: ``bool``, default True,
            Whether to use the novel ``ColumnEmbedding`` layer proposed in the
            ``TabTransformer`` architecture for the categorical features.
            The ``ColumnEmbedding`` layer is an alternative to positional encoding that is applied
            in the Transformers in Natural Langauge Processing application settings.

        encode_categorical_values: ``bool``, default True,
            Whether to (label) encode categorical values,
            If you've already encoded the categorical values using for instance
            Label/Ordinal encoding, you should set this to False,
            otherwise leave it as True.
            In the case of True, categorical values will be mapped to integer indices
            using keras's ``IntegerLookup`` layer.

        numerical_normalization: ``str`` or ``keras.layes.Layer``, default "layer",
            Normalization to use for the numerical features.
    """
    def __init__(self,
                 num_classes: int = 2,
                 head_units_values: UnitsValuesType = (64, 32),
                 input_dim: int = None,
                 features_metadata: dict = None,
                 embedding_dim: int = TabTransformerConfig.embedding_dim,
                 num_transformer_layers: int = TabTransformerConfig.num_transformer_layers,
                 num_attention_heads: int = TabTransformerConfig.num_attention_heads,
                 attention_dropout: float = TabTransformerConfig.attention_dropout,
                 feedforward_dropout: float = TabTransformerConfig.feedforward_dropout,
                 feedforward_multiplier: int = TabTransformerConfig.feedforward_multiplier,
                 norm_epsilon: float = TabTransformerConfig.norm_epsilon,
                 use_column_embedding: bool = TabTransformerConfig.use_column_embedding,
                 encode_categorical_values: bool = TabTransformerConfig.encode_categorical_values,
                 numerical_normalization: NormalizationType = TabTransformerConfig.numerical_normalization,
                 **kwargs
                 ):
        # Since the parent `TabTransformer` class subclasses its LayerFlow version
        # which accepts layers as input and does expose a head argumnet to pass the layer.
        # we can create a relevant head layer, in this case, the `ClassificationHead` layer
        # and pass its instance to the parent who will pass it along to its parent as part of
        # the **kwargs dict
        head = ClassificationHead(num_classes=num_classes,
                                  units_values=head_units_values,
                                  activation_hidden="relu",
                                  normalization="batch",
                                  name="tabtransformer_classifier_head")
        super().__init__(input_dim=input_dim,
                         features_metadata=features_metadata,
                         embedding_dim=embedding_dim,
                         num_transformer_layers=num_transformer_layers,
                         num_attention_heads=num_attention_heads,
                         attention_dropout=attention_dropout,
                         feedforward_dropout=feedforward_dropout,
                         feedforward_multiplier=feedforward_multiplier,
                         norm_epsilon=norm_epsilon,
                         use_column_embedding=use_column_embedding,
                         encode_categorical_values=encode_categorical_values,
                         numerical_normalization=numerical_normalization,
                         head=head,
                         **kwargs)
        self.num_classes = num_classes
        self.head_units_values = head_units_values

    def get_config(self):
        config = super().get_config()
        config.update({'num_classes': self.num_classes,
                       'head_units_values': self.head_units_values
                       })
        return config


class TabTransformerRegressor(TabTransformer):
    """
    TabTransformerRegressor based on the TabTransformer architecture
    as proposed by Xin Huang et al. in the paper,
    TabTransformer: Tabular Data Modeling Using Contextual Embeddings.

    TabTransformer, a novel deep tabular data modeling architecture for
    supervised and semi-supervised learning.
    The TabTransformer is built upon self-attention based Transformers.
    The Transformer layers transform the embeddings of categorical features
    into robust contextual embeddings to achieve higher prediction accuracy.

    Reference(s):
        https://arxiv.org/abs/2012.06678

    Args:
        num_outputs: ``int``, default 1,
            Number of regression outputs to predict.

        head_units_values: ``List[int]`` or ``Tuple[int]``, default [64, 32],
            Units values to use in the hidden layers in the ``TabTRegressionHead``.
            For each value in the list/tuple,
            a hidden layer of that dimensionality is added to the head.

        input_dim: ``int``,
            Dimensionality of the input dataset,
            or the number of features in the dataset.

        features_metadata: ``dict``,
            A nested dictionary of metadata for features where
            categorical sub-dictionary is a mapping of categorical feature names to a tuple of
            feature indices and the lists of unique values (vocabulary) in them,
            while numerical dictionary is a mapping of numerical feature names to their indices.
            `{feature_name: (feature_idx, vocabulary)}` for feature in categorical features.
            `{feature_name: feature_idx}` for feature in numerical features.
            You can get this dictionary from
                >>> from teras.utils import get_features_metadata_for_embedding
                >>> metadata_dict = get_features_metadata_for_embedding(dataframe,
                ..                                                      categorical_features,
                ..                                                      numerical_features)

        embedding_dim: ``int``, default 32,
            Dimensionality of the learnable feature embeddings for categorical features.

        num_transformer_layers: ``int``, default 6,
            Number of transformer layers to use in the ``Encoder`` layer.
            The ``Encoder`` is used to contextualize the learned feature embeddings.

        num_attention_heads: ``int``, default 8,
            Number of attention heads to use in the ``MultiHeadSelfAttention`` layer
            that is part of the ``Transformer`` layer which in turn is part of the
            ``Encoder`` layer.

        attention_dropout: ``float``, default 0.0,
            Dropout rate to use in the ``MultiHeadSelfAttention`` layer in the transformer layer.

        feedforward_dropout: ``float``, default 0.0,
            Dropout rate to use for the ``Dropout`` layer in the ``FeedForward`` layer.

        feedforward_multiplier: ``int``, default 4.
            Multiplier that is multiplied with the ``embedding_dim``
            and the resultant value is used as hidden dimensions value for the
            hidden layer in the ``FeedForward`` layer.

        norm_epsilon: ``float``, default 1e-6,
            A very small number used for normalization in the ``LayerNormalization`` layer.

        use_column_embedding: ``bool``, default True,
            Whether to use the novel ``ColumnEmbedding`` layer proposed in the
            ``TabTransformer`` architecture for the categorical features.
            The ``ColumnEmbedding`` layer is an alternative to positional encoding that is applied
            in the Transformers in Natural Langauge Processing application settings.

        encode_categorical_values: ``bool``, default True,
            Whether to (label) encode categorical values,
            If you've already encoded the categorical values using for instance
            Label/Ordinal encoding, you should set this to False,
            otherwise leave it as True.
            In the case of True, categorical values will be mapped to integer indices
            using keras's ``IntegerLookup`` layer.

        numerical_normalization: ``str`` or ``keras.layes.Layer``, default "layer",
            Normalization to use for the numerical features.
    """
    def __init__(self,
                 num_outputs: int = 1,
                 head_units_values: UnitsValuesType = (64, 32),
                 input_dim: int = None,
                 features_metadata: dict = None,
                 embedding_dim: int = TabTransformerConfig.embedding_dim,
                 num_transformer_layers: int = TabTransformerConfig.num_transformer_layers,
                 num_attention_heads: int = TabTransformerConfig.num_attention_heads,
                 attention_dropout: float = TabTransformerConfig.attention_dropout,
                 feedforward_dropout: float = TabTransformerConfig.feedforward_dropout,
                 feedforward_multiplier: int = TabTransformerConfig.feedforward_multiplier,
                 norm_epsilon: float = TabTransformerConfig.norm_epsilon,
                 use_column_embedding: bool = TabTransformerConfig.use_column_embedding,
                 encode_categorical_values: bool = TabTransformerConfig.encode_categorical_values,
                 numerical_normalization: NormalizationType = TabTransformerConfig.numerical_normalization,
                 **kwargs
                 ):
        head = RegressionHead(num_outputs=num_outputs,
                              units_values=head_units_values,
                              activation_hidden="relu",
                              normalization="batch",
                              name="tabtransformer_regressor_head")
        super().__init__(input_dim=input_dim,
                         features_metadata=features_metadata,
                         embedding_dim=embedding_dim,
                         num_transformer_layers=num_transformer_layers,
                         num_attention_heads=num_attention_heads,
                         attention_dropout=attention_dropout,
                         feedforward_dropout=feedforward_dropout,
                         feedforward_multiplier=feedforward_multiplier,
                         norm_epsilon=norm_epsilon,
                         use_column_embedding=use_column_embedding,
                         encode_categorical_values=encode_categorical_values,
                         numerical_normalization=numerical_normalization,
                         head=head,
                         **kwargs)
        self.num_outputs = num_outputs
        self.head_units_values = head_units_values

    def get_config(self):
        config = super().get_config()
        config.update({'num_outputs': self.num_outputs,
                       'head_units_values': self.head_units_values,
                       })
        return config



import tensorflow as tf
from teras.layers.preprocessing.label_encoding import LabelEncoding
from teras.utils import get_features_metadata_for_embedding
import pandas as pd
import numpy as np
import pytest


@pytest.fixture
def setup_data():
    df = pd.DataFrame({'player_level': [70, 90, 80, 90, 100, 90, 70, 80],
                       'income': [1000, 1270, 900, 1000, 1234, 5678, 9999, 0],
                       'shirt_number': [10, 7, 7, 10, 10, 10, 7, 10],
                       })
    metadata = get_features_metadata_for_embedding(df,
                                                   categorical_features=["player_level", "shirt_number"],
                                                   numerical_features=["income"])
    return df.values, metadata


def test_label_encoding_output_shape_when_concatenate_numerical_features_is_false(setup_data):
    data, metadata = setup_data
    label_encoding = LabelEncoding(features_metadata=metadata,
                                   concatenate_numerical_features=False)
    outputs = label_encoding(data)
    assert len(tf.shape(outputs)) == 2
    assert tf.shape(outputs)[0] == 8
    assert tf.shape(outputs)[1] == 2    # number of categorical features = 2


def test_label_encoding_output_shape_when_concatenate_numerical_features_is_true(setup_data):
    data, metadata = setup_data
    label_encoding = LabelEncoding(features_metadata=metadata,
                                   concatenate_numerical_features=True)
    outputs = label_encoding(data)
    assert len(tf.shape(outputs)) == 2
    assert tf.shape(outputs)[0] == 8
    assert tf.shape(outputs)[1] == 3    # total number of features = 3


def test_label_encoding_keeps_features_order(setup_data):
    data, metadata = setup_data
    label_encoding = LabelEncoding(features_metadata=metadata,
                                   concatenate_numerical_features=True,
                                   keep_features_order=True)
    outputs = label_encoding(data)
    assert len(tf.shape(outputs)) == 2
    assert tf.shape(outputs)[0] == 8
    assert tf.shape(outputs)[1] == 3    # total number of features = 3
    assert all(data[:, 1] == outputs[:, 1])



import tensorflow as tf
from teras.layers.tabnet.tabnet_attentive_transformer import TabNetAttentiveTransformer


def test_tabnet_attentive_transformer_valid_call():
    attentive_transformer = TabNetAttentiveTransformer(data_dim=5)
    inputs = tf.ones((8, 5))
    prior_scales = tf.ones(tf.shape(inputs))
    outputs = attentive_transformer(inputs, prior_scales)


def test_tabnet_attentive_transformer_output_shape():
    attentive_transformer = TabNetAttentiveTransformer(data_dim=5)
    inputs = tf.ones((8, 5))
    prior_scales = tf.ones(tf.shape(inputs))
    outputs = attentive_transformer(inputs, prior_scales)
    assert len(tf.shape(outputs)) == 2
    assert tf.shape(outputs)[0] == 8
    assert tf.shape(outputs)[1] == 5




from tensorflow import keras
from teras.utils import get_normalization_layer
from teras.utils.types import (UnitsValuesType,
                               NormalizationType,
                               ActivationType)


@keras.saving.register_keras_serializable(package="teras.layers.common")
class RegressionHead(keras.layers.Layer):
    """
    Regression Head to use on top of the architectures for regression.

    Args:
        num_outputs: ``int``, default 1,
            Number of regression outputs to predict.

        units_values: ``List[int]`` or ``Tuple[int]``, default (64, 32),
            For each value in the sequence a hidden layer of that dimension
            preceded by a normalization layer (if specified) is
            added to the ``RegressionHead``.

        activation_hidden: ``callable`` or ``str``, default "relu",
            Activation function to use in hidden dense layers.

        normalization: ``keras.layers.Layer`` or ``str``, default "batch",
            Normalization layer to use.
            If specified a normalization layer is applied after each hidden layer.
            If None, no normalization layer is applied.
            You can either pass a keras normalization layer or name for a layer implemented by keras.
    """
    def __init__(self,
                 num_outputs: int = 1,
                 units_values: UnitsValuesType = (64, 32),
                 activation_hidden: ActivationType = "relu",
                 normalization: NormalizationType = "batch",
                 **kwargs):
        super().__init__(**kwargs)
        self.num_outputs = num_outputs
        self.units_values = units_values
        self.activation_hidden = activation_hidden
        self.normalization = normalization

        self.hidden_block = None
        if self.units_values is not None:
            self.hidden_block = keras.models.Sequential(name="inner_head")
            for units in self.units_values:
                if self.normalization is not None:
                    self.hidden_block.add(get_normalization_layer(self.normalization))
                self.hidden_block.add(keras.layers.Dense(units,
                                                         activation=self.activation_hidden))
        self.output_layer = keras.layers.Dense(self.num_outputs)

    def call(self, inputs):
        x = inputs
        if self.hidden_block is not None:
            x = self.hidden_block(x)
        outputs = self.output_layer(x)
        return outputs

    def get_config(self):
        config = super().get_config()

        if isinstance(self.activation_hidden, str):
            activation_hidden_serialized = self.activation_hidden
        else:
            activation_hidden_serialized = keras.layers.serialize(self.activation_hidden)

        if isinstance(self.normalization, str):
            normalization_serialized = self.normalization
        else:
            normalization_serialized = keras.layers.serialize(self.normalization)

        config.update({'num_outputs': self.num_outputs,
                       'units_values': self.units_values,
                       'activation_hidden': activation_hidden_serialized,
                       'normalization': normalization_serialized}
                      )
        return config


@keras.saving.register_keras_serializable(package="teras.layers.common")
class ClassificationHead(keras.layers.Layer):
    """
    Classification head to use on top of the architectures for classification.

    Args:
        num_classes: ``int``, default 2,
            Number of classes to predict.

        units_values: ``List[int]`` or ``Tuple[int]``, default (64, 32),
            For each value in the sequence a hidden layer of that dimension
            preceded by a normalization layer (if specified) is
            added to the ``ClassificationHead``.

        activation_hidden: ``callable``, ``keras.layers.Layer``, or ``str``, default "relu",
            Activation function to use in hidden dense layers.

        activation_out: ``callable``, ``keras.layers.Layer``, or ``str``,
            Activation function to use for the output layer.
            If not specified, `sigmoid` is used for binary and `softmax` is used for
            multiclass classification.

        normalization: ``keras.layers.Layer`` or ``str``, default "batch",
            Normalization layer to use.
            If specified a normalization layer is applied after each hidden layer.
            If None, no normalization layer is applied.
            You can either pass a keras normalization layer or name for a layer implemented by keras.
    """
    def __init__(self,
                 num_classes: int = 2,
                 units_values: UnitsValuesType = (64, 32),
                 activation_hidden: ActivationType = "relu",
                 activation_out: ActivationType = None,
                 normalization: NormalizationType = "batch",
                 **kwargs):
        super().__init__(**kwargs)
        self.num_classes = 1 if num_classes <= 2 else num_classes
        self.units_values = units_values
        self.activation_hidden = activation_hidden
        self.activation_out = activation_out
        if self.activation_out is None:
            self.activation_out = "sigmoid" if self.num_classes == 1 else "softmax"
        self.normalization = normalization

        self.hidden_block = None
        if self.units_values is not None:
            self.hidden_block = keras.models.Sequential(name="inner_head")
            for units in self.units_values:
                if self.normalization is not None:
                    self.hidden_block.add(get_normalization_layer(self.normalization))
                self.hidden_block.add(keras.layers.Dense(units,
                                                         activation=self.activation_hidden))
        self.output_layer = keras.layers.Dense(self.num_classes,
                                               activation=self.activation_out)

    def call(self, inputs):
        x = inputs
        if self.hidden_block is not None:
            x = self.hidden_block(x)
        outputs = self.output_layer(x)
        return outputs

    def get_config(self):
        config = super().get_config()

        if isinstance(self.activation_hidden, str):
            activation_hidden_serialized = self.activation_hidden
        else:
            activation_hidden_serialized = keras.layers.serialize(self.activation_hidden)

        if isinstance(self.activation_out, str):
            activation_out_serialized = self.activation_out
        else:
            activation_out_serialized = keras.layers.serialize(self.activation_out)

        if isinstance(self.normalization, str):
            normalization_serialized = self.normalization
        else:
            normalization_serialized = keras.layers.serialize(self.normalization)

        config.update({'num_classes': self.num_classes,
                       'units_values': self.units_values,
                       'activation_hidden': activation_hidden_serialized,
                       'activation_out': activation_out_serialized,
                       'normalization': normalization_serialized}
                      )
        return config



import tensorflow as tf
from teras.layers.rtdl_resnet.rtdl_resnet_block import RTDLResNetBlock


def test_rtdl_resnet_block_valid_call():
    resnet_block = RTDLResNetBlock(units=64)
    inputs = tf.ones((8, 4), dtype=tf.float32)
    outputs = resnet_block(inputs)


def test_rtdl_resnet_block_output_shape():
    resnet_block = RTDLResNetBlock(units=64)
    inputs = tf.ones((8, 4), dtype=tf.float32)
    outputs = resnet_block(inputs)
    assert len(tf.shape(outputs)) == 2
    assert tf.shape(outputs)[0] == 8
    assert tf.shape(outputs)[1] == 4



import tensorflow as tf
from teras.layers.saint.saint_projection_head import SAINTProjectionHead


def test_saint_projection_head_output_shape():
    saint_transformer = SAINTProjectionHead(hidden_dim=16,
                                            output_dim=8)
    inputs = tf.ones((16, 10), dtype=tf.float32)
    outputs = saint_transformer(inputs)
    assert len(tf.shape(outputs)) == 2
    assert tf.shape(outputs)[0] == 16
    assert tf.shape(outputs)[1] == 8



from tensorflow import keras
from teras.layerflow.layers.saint.saint_projection_head import SAINTProjectionHead as _SAINTProjectionHeadLF


@keras.saving.register_keras_serializable(package="teras.layers.saint")
class SAINTProjectionHead(_SAINTProjectionHeadLF):
    """
    ProjectionHead layer that is used in the contrastive learning phase of
    the SAINTPretrainer to project embeddings to a lower dimension.
    According to the SAINT paper,
    "The use of a projection head to reduce dimensionality before computing
    contrastive loss is common in vision and indeed also improves results
    on tabular data."

    Reference(s):
    https://arxiv.org/abs/2106.01342

    Args:
        hidden_dim: ``int``, default 64,
            Dimensionality of the hidden layer.
            In the official implementation, it is computed as follows,
            `hidden_dim = 6 * embedding_dim * number_of_features // 5`

        hidden_activation: default "relu":
            Activation function to use in the hidden layer.

        output_dim: ``int``, default 32,
            Dimensionality of the output layer.
            In the official implementation, it is computed as follows,
            `output_dim = embedding_dim * number_of_features // 5`
    """
    def __init__(self,
                 hidden_dim: int = 64,
                 hidden_activation="relu",
                 output_dim: int = 32,
                 **kwargs):
        hidden_block = keras.layers.Dense(units=hidden_dim,
                                          activation=hidden_activation)
        output_layer = keras.layers.Dense(units=output_dim)

        super().__init__(hidden_block=hidden_block,
                         output_layer=output_layer,
                         **kwargs)
        self.hidden_dim = hidden_dim
        self.hidden_activation = hidden_activation
        self.output_dim = output_dim

    def get_config(self):
        config = {'name': self.name,
                  'trainable': self.trainable,
                  'hidden_dim': self.hidden_dim,
                  'hidden_activation': self.hidden_activation,
                  'output_dim': self.output_dim}
        return config

    # need to override the from_config since the layerflow version is parent
    # and it tries to pop out layers to deserialize which in this default
    # api case don't exist in the configuration and hence causes KeyError
    @classmethod
    def from_config(cls, config):
        return cls(**config)



from teras.layers.common.head import ClassificationHead, RegressionHead
import tensorflow as tf


def test_classification_head_binary_classification_output_shape():
    head = ClassificationHead(num_classes=2)
    inputs = tf.ones(shape=(128, 16), dtype=tf.float32)
    outputs = head(inputs)
    assert tf.shape(outputs)[0] == 128
    assert tf.shape(outputs)[1] == 1


def test_classification_head_multiclass_classification_output_shape():
    head = ClassificationHead(num_classes=3)
    inputs = tf.ones(shape=(128, 16), dtype=tf.float32)
    outputs = head(inputs)
    assert tf.shape(outputs)[0] == 128
    assert tf.shape(outputs)[1] == 3


def test_regression_head_single_regression_output_shape():
    head = RegressionHead(num_outputs=1)
    inputs = tf.ones(shape=(128, 16), dtype=tf.float32)
    outputs = head(inputs)
    assert len(tf.shape(outputs)) == 2
    assert tf.shape(outputs)[0] == 128
    assert tf.shape(outputs)[1] == 1


def test_regression_head_multi_regression_output_shape():
    head = RegressionHead(num_outputs=3)
    inputs = tf.ones(shape=(128, 16), dtype=tf.float32)
    outputs = head(inputs)
    assert len(tf.shape(outputs)) == 2
    assert tf.shape(outputs)[0] == 128
    assert tf.shape(outputs)[1] == 3





import tensorflow as tf
from teras.layers.saint.saint_encoder import SAINTEncoder


def test_saint_encoder_output_shape():
    encoder = SAINTEncoder(data_dim=10,
                           embedding_dim=32)
    inputs = tf.ones((16, 10, 32), dtype=tf.float32)
    outputs = encoder(inputs)
    assert len(tf.shape(outputs)) == 3
    assert tf.shape(outputs)[0] == 16
    assert tf.shape(outputs)[1] == 10
    assert tf.shape(outputs)[2] == 32



import tensorflow as tf
from tensorflow import keras
from teras.layerflow.models.tabtransformer import (TabTransformer,
                                                   TabTransformerPretrainer)
from teras.layers import TabTransformerColumnEmbedding
from teras.layers import CategoricalFeatureEmbedding, Encoder
from teras.utils import get_features_metadata_for_embedding
import pandas as pd
import numpy as np
import os


class TabTransformerTest(tf.test.TestCase):
    def setUp(self):
        data = pd.DataFrame({"income": np.ones(10),
                             "goals": np.ones(10),
                             'player_level': [5, 7, 9, 8, 9, 10, 9, 7, 8, 9],
                             'shirt_number': [7, 10, 10, 7, 7, 10, 10, 10, 7, 10]})
        categorical_feats = ["player_level", "shirt_number"]
        numerical_feats = ["income", "goals"]
        self.features_metadata = get_features_metadata_for_embedding(pd.DataFrame(data),
                                                                     categorical_features=categorical_feats,
                                                                     numerical_features=numerical_feats)
        self.categorical_feature_embedding = CategoricalFeatureEmbedding(features_metadata=self.features_metadata,
                                                                         embedding_dim=32)
        self.numerical_feature_normalization = keras.layers.LayerNormalization()
        self.column_embedding = TabTransformerColumnEmbedding(num_categorical_features=2)
        self.encoder = Encoder()
        self.head = keras.layers.Dense(1)
        self.data_batch = data.values

    def test_valid_call(self):
        model = TabTransformer(input_dim=4,
                               categorical_feature_embedding=self.categorical_feature_embedding,
                               column_embedding=self.column_embedding,
                               numerical_feature_normalization=self.numerical_feature_normalization,
                               encoder=self.encoder,
                               head=self.head)
        model(self.data_batch)

    def test_save_and_load(self):
        model = TabTransformer(input_dim=4,
                               categorical_feature_embedding=self.categorical_feature_embedding,
                               column_embedding=self.column_embedding,
                               numerical_feature_normalization=self.numerical_feature_normalization,
                               encoder=self.encoder,
                               head=self.head)
        save_path = os.path.join(self.get_temp_dir(), "tabtransformer_lf.keras")
        model.save(save_path, save_format="keras_v3")
        reloaded_model = keras.models.load_model(save_path)
        outputs_original = model(self.data_batch)
        outputs_reloaded = reloaded_model(self.data_batch)
        self.assertAllClose(outputs_original, outputs_reloaded)


class TabTransformerPretrainerTest(tf.test.TestCase):
    def setUp(self):
        data = pd.DataFrame({"income": np.ones(10),
                             "goals": np.ones(10),
                             'player_level': [5, 7, 9, 8, 9, 10, 9, 7, 8, 9],
                             'shirt_number': [7, 10, 10, 7, 7, 10, 10, 10, 7, 10]})
        categorical_feats = ["player_level", "shirt_number"]
        numerical_feats = ["income", "goals"]
        self.features_metadata = get_features_metadata_for_embedding(pd.DataFrame(data),
                                                                     categorical_features=categorical_feats,
                                                                     numerical_features=numerical_feats)
        self.categorical_feature_embedding = CategoricalFeatureEmbedding(features_metadata=self.features_metadata,
                                                                         embedding_dim=32)
        self.numerical_feature_normalization = keras.layers.LayerNormalization()
        self.column_embedding = TabTransformerColumnEmbedding(num_categorical_features=2)
        self.encoder = Encoder()
        self.data_batch = data.values
        self.num_features = 4
        self.num_features_to_replace = 2
        feature_indices_to_replace = tf.random.uniform((tf.shape(self.data_batch)[0], self.num_features_to_replace),
                                                       maxval=self.num_features,
                                                       dtype=tf.int32)
        self.mask = tf.reduce_max(tf.one_hot(feature_indices_to_replace,
                                             depth=self.num_features),
                                  axis=1)

    def test_valid_call(self):
        base_model = TabTransformer(input_dim=4,
                                    categorical_feature_embedding=self.categorical_feature_embedding,
                                    column_embedding=self.column_embedding,
                                    numerical_feature_normalization=self.numerical_feature_normalization,
                                    encoder=self.encoder)
        pretrainer = TabTransformerPretrainer(model=base_model,
                                              features_metadata=self.features_metadata)
        pretrainer(self.data_batch, mask=self.mask)

    def test_save_and_load(self):
        base_model = TabTransformer(input_dim=4,
                                    categorical_feature_embedding=self.categorical_feature_embedding,
                                    column_embedding=self.column_embedding,
                                    numerical_feature_normalization=self.numerical_feature_normalization,
                                    encoder=self.encoder)
        pretrainer = TabTransformerPretrainer(model=base_model,
                                              features_metadata=self.features_metadata)
        save_path = os.path.join(self.get_temp_dir(), "tabtransformer_pretrainer_lf.keras")
        pretrainer.save(save_path, save_format="keras_v3")
        reloaded_model = keras.models.load_model(save_path)
        outputs_original = pretrainer(self.data_batch, mask=self.mask)
        outputs_reloaded = reloaded_model(self.data_batch, mask=self.mask)
        # We can't check for AllClose because the call method randomly shuffles the inputs



from tensorflow import keras
import tensorflow as tf
from tensorflow.keras import layers, models
from typing import List, Union, Tuple
import pandas as pd
import numpy as np
from warnings import warn
from teras.utils.types import ActivationType
from teras import activations

LayerType = Union[str, layers.Layer]
LAYERS_COLLECTION = Union[List[layers.Layer], layers.Layer, models.Model]
LAYERS_CONFIGS_COLLECTION = Union[List[dict], dict]


def tf_random_choice(inputs,
                     n_samples: int,
                     p: List[float] = None):
    """Tensorflow equivalent of np.random.choice

    Args:
        inputs: Tensor to sample from
        n_samples: Number of samples to draw
        p: Probabilities for each sample

    Reference(s):
        https://stackoverflow.com/questions/41123879/numpy-random-choice-in-tensorflow
    """
    probs = p
    if probs is None:
        probs = tf.random.uniform([n_samples])
    # Normalize probabilities so they all sum to one
    probs /= tf.reduce_sum(probs)
    probs = tf.expand_dims(probs, 0)
    # Convert them to log probabilities because that's what the tf.random.categorical function expects
    probs = tf.math.log(probs)
    indices = tf.random.categorical(probs, n_samples, dtype=tf.int32)
    return tf.gather(inputs, indices)


def get_normalization_layer(normalization: LayerType) -> layers.Layer:
    """
    Creates and returns a keras normalization layer if not already.

    Args:
        normalization: default BatchNormalization.
            If type of normalization is a keras layer, it is returned as is.
            If it is of type, str, that is, it is a name,
            then relevant normalization layer is returned

    Returns:
        Keras Normalization Layer
    """
    if isinstance(normalization, str):
        normalization = normalization.lower()
        if normalization in ["batchnormalization", "batch"]:
            normalization_layer = layers.BatchNormalization()
        elif normalization in ["layernormalization", "layer"]:
            normalization_layer = layers.LayerNormalization()
        elif normalization in ["unitnormalization", "unit"]:
            normalization_layer = layers.UnitNormalization()
        elif normalization in ["groupnormalization", "group"]:
            normalization_layer = layers.GroupNormalization()
        else:
            raise ValueError(f"Normalization type {normalization} is not supported in Keras.")
    elif isinstance(normalization, layers.Layer):
        normalization_layer = normalization
    else:
        raise ValueError(f"Invalid Normalization value type. Expected type str or keras.layers.Layer"
                         f" Received value {normalization} of type {type(normalization)}")
    return normalization_layer


def get_activation(activation: ActivationType):
    """
    Retrieves and returns a keras activation function if not already.

    Args:
        activation:
            If type of activation is a keras function, it is returned as is.
            If it is of type, str, that is, it is a name,
            then relevant activation function is returned, if it is offered
            by either ``Teras`` or ``Keras``, otherwise an error is raised.

    Returns:
        Keras Activation function
    """
    if isinstance(activation, str):
        activation = activation.lower()
        # First check if Keras offers that function
        try:
            activation_func = keras.activations.get(activation)
        except ValueError:
            # Then check if Teras offers it
            if activation == "glu":
                activation_func = activations.glu
            elif activation == "geglu":
                activation_func = activations.geglu
            elif activation == "gumblesoftmax":
                activation_func = activations.gumbel_softmax
            elif activation == "sparsemax":
                activation_func = activations.sparsemax
            elif activation == "sparsemoid":
                from teras.utils.node import sparsemoid
                activation_func = sparsemoid
            else:
                # Otherwise return error
                raise ValueError(f"{activation} function's name is either incorrect "
                                 f"or this activation is not offered by either Keras and Teras. ")
    elif callable(activation):
        activation_func = activation

    else:
        TypeError("Unsupported type for `activation` argument. "
                  f"Expected type(s): [`str`, `callable`]. Received: {type(activation)}")
    return activation_func


def get_initializer(initializer: LayerType):
    """
    Retrieves and returns a keras initializer function if not already.

    Args:
        initializer: default uniform.
            If type of initializer is a keras function, it is returned as is.
            If it is of type, str, that is, it is a name,
            then relevant initializer function is returned

    Returns:
        Keras initializer function
    """
    if isinstance(initializer, str):
        initializer = initializer.lower()
        try:
            initializer_func = keras.initializers.get(initializer)
        except ValueError:
            raise ValueError(f"{initializer} function's name is either wrong or this activation is not supported by Keras."
                             f"Please contact Teras team, and we'll make sure to add this to Teras.")
    else:
        initializer_func = initializer
    return initializer_func


def get_categorical_features_cardinalities(dataframe,
                                           categorical_features):
    cardinalities = []
    for feature in categorical_features:
        cardinalities.append(dataframe[feature].nunique())
    return cardinalities


def get_features_metadata_for_embedding(dataframe: pd.DataFrame,
                                        categorical_features=None,
                                        numerical_features=None):
    """
    Utility function that create metadata for features in a given dataframe
    required by the Categorical and Numerical embedding layers in Teras.
    For numerical features, it maps each feature name to feature index.
    For categorical features, it maps each feature name to a tuple of
    feature index and vocabulary of words in that categorical feature.
    This metadata is usually required by the architectures that create embeddings
    of Numerical or Categorical features,
    such as TabTransformer, TabNet, FT-Transformer, etc.

    Args:
        dataframe: Input dataframe
        categorical_features: List of names of categorical features in the input dataset
        numerical_features: List of names of categorical features in the input dataset

    Returns:
        A dictionary which contains sub-dictionaries for categorical and numerical features
        where categorical dictionary is a mapping of categorical feature names to a tuple of
        feature indices and the lists of unique values (vocabulary) in them,
        while numerical dictionary is a mapping of numerical feature names to their indices.
        {feature_name: (feature_idx, vocabulary)} for feature in categorical features.
        {feature_name: feature_idx} for feature in numerical features.
    """
    if categorical_features is None and numerical_features is None:
        raise ValueError("Both `categorical_features` and `numerical_features` cannot be None at the same time. "
                         "You must pass value for at least one of them. If your dataset contains both types of "
                         "features then it is strongly recommended to pass features names for both types. "
                         f"Received, `categorical_features`: {categorical_features}, "
                         f"`numerical_features`: {numerical_features}")
    features_meta_data = {}
    categorical_features_metadata = {}
    numerical_features_metadata = {}
    for idx, col in enumerate(dataframe.columns):
        if categorical_features is not None:
            if col in categorical_features:
                vocabulary = sorted(list(dataframe[col].unique()))
                categorical_features_metadata.update({col: (idx, vocabulary)})
        if numerical_features is not None:
            if col in numerical_features:
                numerical_features_metadata.update({col: idx})

    features_meta_data["categorical"] = categorical_features_metadata
    features_meta_data["numerical"] = numerical_features_metadata
    return features_meta_data


def dataframe_to_tf_dataset(
        dataframe: pd.DataFrame,
        target: Union[str, list] = None,
        shuffle: bool = True,
        batch_size: int = 1024,
):
    """
    Builds a tf.data.Dataset from a given pandas dataframe

    Args:
        dataframe: `pd.DataFrame`,
            A pandas dataframe
        target: `str` or `list`,
            Name of the target column or list of names of the target columns.
        shuffle: `bool`, default True
            Whether to shuffle the dataset
        batch_size: `int`, default 1024,
            Batch size

    Returns:
         A tf.data.Dataset dataset
    """
    df = dataframe.copy()
    if target is not None:
        if isinstance(target, (list, tuple, set)):
            labels = []
            for feat in target:
                labels.append(df.pop(feat).values)
            labels = tf.transpose(tf.constant(labels))
        else:
            labels = df.pop(target)
            labels = labels.values
        dataset = tf.data.Dataset.from_tensor_slices((df.values, labels))
    else:
        dataset = tf.data.Dataset.from_tensor_slices(df.values)
    if shuffle:
        dataset = dataset.shuffle(buffer_size=len(dataframe))
    dataset = dataset.batch(batch_size)
    dataset = dataset.prefetch(batch_size)
    return dataset


def convert_dict_to_array_tensor(dict_tensor):
    """
    Converts a batch of data taken from tensorflow dictionary format dataset
    to array format.
    Args:
        dict_tensor: A batch of data taken from tensorflow dictionary format dataset.

    Returns:
        Array format data.
    """
    if not isinstance(dict_tensor, dict):
        warn("Given tensor is not in dictionary format. Hence no processing will be applied. \n"
             f"Expected type: {dict}, Received type: {type(dict_tensor)}")
        return

    feature_names = dict_tensor.keys()
    array_tensor = [tf.expand_dims(dict_tensor[feature_name], axis=1)
                    for feature_name in feature_names]
    array_tensor = tf.concat(array_tensor, axis=1)
    return array_tensor


def serialize_layers_collection(layers_collection: LAYERS_COLLECTION):
    """
    Serializes a collection of keras layers/models.

    Returns:
        If layers_collection is an instance of list or tuple, it returns a list of serialized layers
        otherwise it simply returns the serialized layer or model.
    """
    if isinstance(layers_collection, (list, tuple)):
        layers_collection_serialized = [keras.layers.serialize(layer)
                                        for layer in layers_collection]
    else:
        # We assume it's either of type Layer or Model, or None
        layers_collection_serialized = keras.layers.serialize(layers_collection)

    return layers_collection_serialized


def deserialize_layers_collection(layers_configs_collection: LAYERS_CONFIGS_COLLECTION):
    """
    De-serializes a collection of keras layers/models configs.

    Returns:
        If layers_configs_collection is a list of layers config dictionaries, it returns a
        list of de-serialized layers otherwise if it's a config dictionary, it returns the
         de-serialized layer or model.
    """
    if isinstance(layers_configs_collection, list):
        layers_collection_deserialized = [keras.layers.deserialize(layer)
                                          for layer in layers_configs_collection]
    else:
        # We assume it's either a dict or None
        layers_collection_deserialized = keras.layers.deserialize(layers_configs_collection)

    return layers_collection_deserialized


def inject_missing_values(x: pd.DataFrame,
                          miss_rate=0.1,
                          return_dataframe=True):
    """
    Injects missing (np.nan) values in the given dataset.

    Args:
        x: A pandas dataframe.
        miss_rate: The fraction of missing values that should be introduced.
            Should be between 0-1. Defaults to 0.1
        return_dataframe: If False, numpy ndarray will be returned,
            otherwise pandas dataframe will be returned.

    Returns:
        Data with missing values.

    Example:
        ```python
        data = np.arange(1000).reshape(50, 20)
        data = inject_missing_values(data, miss_rate=0.2, return_dataframe=False)
        ```
    """
    x_with_missing_data = x.copy()
    is_dataframe = isinstance(x_with_missing_data, pd.DataFrame)

    if is_dataframe:
        x_with_missing_data = x_with_missing_data.values
        is_dataframe = True

    mask = np.random.binomial(1, 1-miss_rate, size=x.shape)
    x_with_missing_data[mask == 0] = np.nan

    if return_dataframe:
        x_with_missing_data = pd.DataFrame(x_with_missing_data,
                                           columns=x.columns if is_dataframe else None)
    return x_with_missing_data


def generate_fake_gemstone_data(num_samples: int = 16):
    """
    Generate fake gemstone like data of specified num_samples.

    Args:
        num_samples:
            Number of samples to generate

    Returns:
        A pandas DataFrame of fake gemstone like data.
    """
    fake_gem_df = pd.DataFrame({
        "cut": np.random.randint(low=0, high=3, size=(num_samples,)),
        "color": np.random.randint(low=0, high=5, size=(num_samples,)),
        "clarity": np.random.randint(low=0, high=4, size=(num_samples,)),
        "depth": np.random.randint(low=0, high=100, size=(num_samples,)),
        "table": np.random.randint(low=0, high=100, size=(num_samples,))
    })
    fake_gem_df = fake_gem_df.astype(np.float32)
    return fake_gem_df



from abc import abstractmethod
import pandas as pd


class BaseDataTransformer:
    """
    Base class for ``DataTransformer`` classes.
    It provides the common methods and attributes.
    """
    def __init__(self):
        self.metadata = dict()
        self.metadata["numerical"] = dict()
        self.metadata["categorical"] = dict()
        self.fitted = False

    @abstractmethod
    def fit(self, x, **kwargs):
        pass

    @abstractmethod
    def transform(self, x, **kwargs):
        pass

    def get_metadata(self):
        """
        Returns:
            named tuple of features metadata.
        """
        return self.metadata

    def fit_transform(self,
                      x: pd.DataFrame,
                      return_dataframe: bool = True):
        self.fit(x)
        return self.transform(x, return_dataframe=return_dataframe)



"""
This module contains general purpose common layers that can't be
classified into some category or attributed to a base architecture
like transformers.
"""
from tensorflow import keras
from tensorflow.keras import layers


@keras.saving.register_keras_serializable(package="teras.layerflow.layers")
class HiLOL(layers.Layer):
    """
    HiLOL, short for, Hidden Layer Output Layer, is a simple
    layer that is — as the name implies — made up of a
    Hidden Layer and an Output Layer.

    This serves as base to many layers that are specific to
    various model architectures but are made up of this very
    simple architecture,
    for instance, the SAINT's ``ProjectionHead`` layer,
    and almost all LayerFlow versions of Head layers

    This helps us save a lot of code duplication, such as
    implementation of ``call()``, ``get_config()``, and
    ``from_config()`` methods.
    """
    def __init__(self,
                 hidden_block: layers.Layer,
                 output_layer: layers.Layer,
                 **kwargs):
        super().__init__(**kwargs)
        self.hidden_block = hidden_block
        self.output_layer = output_layer

    def call(self, inputs):
        x = self.hidden_block(inputs)
        return self.output_layer(x)

    def get_config(self):
        config = super().get_config()
        config.update({'hidden_block': keras.layers.serialize(self.hidden_block),
                       'output_layer': keras.layers.serialize(self.output_layer)
                       })
        return config

    @classmethod
    def from_config(cls, config):
        hidden_block = keras.layers.deserialize(config.pop("hidden_block"))
        output_layer = keras.layers.deserialize(config.pop("output_layer"))
        return cls(hidden_block, output_layer, **config)


import tensorflow as tf
from tensorflow import keras
from teras.models.pcgain import PCGAINGenerator, PCGAINDiscriminator, PCGAINClassifier, PCGAIN
from teras.preprocessing.pcgain import PCGAINDataSampler, PCGAINDataTransformer
from teras.utils import inject_missing_values, generate_fake_gemstone_data
import os
import pytest


class PCGAINGeneratorTest(tf.test.TestCase):
    def setUp(self):
        self.latent_inputs = tf.random.normal((8, 16))

    def test_valid_call(self):
        generator = PCGAINGenerator(data_dim=7)
        outputs = generator(self.latent_inputs)

    def test_output_shape(self):
        generator = PCGAINGenerator(data_dim=7)
        outputs = generator(self.latent_inputs)
        self.assertAllEqual(tf.shape(outputs), (8, 7))

    def test_save_and_load(self):
        generator = PCGAINGenerator(data_dim=7)
        save_path = os.path.join(self.get_temp_dir(), "pcgain_generator.keras")
        generator.save(save_path, save_format="keras_v3")
        reloaded_model = keras.models.load_model(save_path)
        outputs_original = generator(self.latent_inputs)
        outputs_reloaded = reloaded_model(self.latent_inputs)
        # self.assertAllClose(outputs_original, outputs_reloaded)


class GAINDiscriminatorTest(tf.test.TestCase):
    def setUp(self):
        self.inputs = tf.ones((8, 7))

    def test_valid_call(self):
        discriminator = PCGAINDiscriminator(data_dim=7)
        outputs = discriminator(self.inputs)

    def test_output_shape(self):
        discriminator = PCGAINDiscriminator(data_dim=7)
        outputs = discriminator(self.inputs)
        self.assertAllEqual(tf.shape(outputs), (8, 7))

    def test_save_and_load(self):
        discriminator = PCGAINDiscriminator(data_dim=7)
        save_path = os.path.join(self.get_temp_dir(), "pcgain_discriminator.keras")
        discriminator.save(save_path, save_format="keras_v3")
        reloaded_model = keras.models.load_model(save_path)
        outputs_original = discriminator(self.inputs)
        outputs_reloaded = reloaded_model(self.inputs)
        # self.assertAllClose(outputs_original, outputs_reloaded)


class GAINTest(tf.test.TestCase):
    def setUp(self):
        fake_gem_df = generate_fake_gemstone_data(num_samples=48)
        numerical_cols = ["depth", "table"]
        categorical_cols = ["cut", "color", "clarity"]
        x_with_missing = inject_missing_values(fake_gem_df)
        self.data_transformer = PCGAINDataTransformer(numerical_features=numerical_cols,
                                                      categorical_features=categorical_cols)
        x_transformed = self.data_transformer.fit_transform(x_with_missing, return_dataframe=True)
        self.x_train = x_transformed[:32]
        self.x_test = x_transformed[32:]
        self.data_sampler = PCGAINDataSampler()
        self.x_train = self.data_sampler.get_dataset(self.x_train)
        self.x_pretrain = self.data_sampler.get_pretraining_dataset(x_transformed, pretraining_size=0.3)
        self.pcgain_imputer = PCGAIN(input_dim=self.data_sampler.data_dim)
        self.pcgain_imputer.compile()

    def test_valid_call(self):
        # GAIN's call method is just calls the generator's call method
        z = tf.random.normal((8, 16))
        pcgain = PCGAIN(input_dim=7)
        outputs = pcgain(z)

    def test_output_shape(self):
        inputs = tf.random.normal((8, 16))
        pcgain = PCGAIN(input_dim=7)
        outputs = pcgain(inputs)
        self.assertAllEqual(tf.shape(outputs), (8, 7))

    def test_save_and_load(self):
        inputs = tf.random.normal((8, 16))
        pcgain = PCGAIN(input_dim=7)
        save_path = os.path.join(self.get_temp_dir(), "pcgain.keras")
        pcgain.save(save_path, save_format="keras_v3")
        reloaded_model = keras.models.load_model(save_path)
        outputs_original = pcgain(inputs)
        outputs_reloaded = reloaded_model(inputs)
        # self.assertAllClose(outputs_original, outputs_reloaded)

    def test_fit_raises_error_if_not_pretrained_first(self):
        with pytest.raises(AssertionError):
            self.pcgain_imputer.fit(self.x_train)

    def test_fit(self):
        self.pcgain_imputer.compile()
        pretrainer_fit_kwargs = {"epochs": 2}
        classifier_fit_kwargs = {"epochs": 2}
        self.pcgain_imputer.pretrain(self.x_pretrain, pretrainer_fit_kwargs, classifier_fit_kwargs)
        self.pcgain_imputer.fit(self.x_train)

    def test_impute(self):
        imputed_data = self.pcgain_imputer.impute(self.x_test, data_transformer=self.data_transformer)
        assert imputed_data.isna().sum().sum() == 0



import tensorflow as tf
from teras.layers.tabnet.tabnet_feature_transformer_block import TabNetFeatureTransformerBlock


def test_tabnet_feature_transformer_block_valid_call():
    # The first TabNetFeatureTransformerBlock must have
    # use_residual_normalization = False
    feature_transformer = TabNetFeatureTransformerBlock(units=32,
                                                        use_residual_normalization=False)
    inputs = tf.ones((8, 10))
    outputs = feature_transformer(inputs)


def test_tabnet_feature_transformer_block_output_shape():
    feature_transformer = TabNetFeatureTransformerBlock(units=32,
                                                        use_residual_normalization=False)
    inputs = tf.ones((8, 10))
    outputs = feature_transformer(inputs)
    assert len(tf.shape(outputs)) == 2
    assert tf.shape(outputs)[0] == 8
    assert tf.shape(outputs)[1] == 32



from teras.config.base import BaseConfig


class SAINTConfig(BaseConfig):
    """Config class for SAINT architecture to hold the
    default values for the common parameters of SAINT
    architecture in one place."""
    embedding_dim: int = 32
    numerical_embedding_hidden_dim: int = 16
    num_transformer_layers: int = 6
    num_attention_heads: int = 8
    num_inter_sample_attention_heads: int = 8
    attention_dropout: float = 0.1
    inter_sample_attention_dropout: float = 0.1
    feedforward_dropout: float = 0.1
    feedforward_multiplier: int = 4
    norm_epsilon: float = 1e-6
    encode_categorical_values: bool = True
    apply_attention_to_features: bool = True
    apply_attention_to_rows: bool = True



import tensorflow as tf
from tensorflow import keras
from teras.layers import (VimeEncoder,
                          VimeMaskEstimator,
                          VimeFeatureEstimator,
                          VimePredictor,
                          VimeMaskGenerationAndCorruption)
from teras.utils.types import ActivationType
from teras.layerflow.models.vime import (VimeSelf as _VimeSelfLF,
                                         VimeSemi as _VimeSemiLF)


@keras.saving.register_keras_serializable(package="teras.models")
class VimeSelf(_VimeSelfLF):
    """Self-supervised learning part in the paper
    "VIME: Extending the Success of Self- and
    Semi-supervised Learning to Tabular Domain"
    by Jinsung Yoon et al.

    Reference(s):
        https://proceedings.neurips.cc/paper/2020/hash/7d97667a3e056acab9aaf653807b4a03-Abstract.html

    Args:
        encoder_activation: default "relu",
            Activation to use for encoder

        feature_estimator_activation: default "sigmoid",
            Activation to use for feature estimator

        mask_estimator_activation: default "sigmoid",
            Activation to use for mask estimator
    """
    def __init__(self,
                 input_dim: int,
                 encoder_activation: ActivationType = "relu",
                 feature_estimator_activation: ActivationType = "sigmoid",
                 mask_estimator_activation: ActivationType = "sigmoid",
                 **kwargs
                 ):
        encoder = VimeEncoder(input_dim,
                              input_shape=(input_dim,),
                              activation=encoder_activation,
                              name="encoder")
        # Mask estimator
        mask_out = VimeMaskEstimator(input_dim,
                                     activation=mask_estimator_activation,
                                     name="mask_estimator")
        # Feature estimator
        feature_out = VimeFeatureEstimator(input_dim,
                                           activation=feature_estimator_activation,
                                           name="feature_estimator")
        super().__init__(**kwargs)
        self.encoder_activation = encoder_activation
        self.feature_estimator_activation = feature_estimator_activation
        self.mask_estimator_activation = mask_estimator_activation

    def build(self, input_shape):
        """Builds layers and functional model"""
        _, input_dim = input_shape
        inputs = keras.layers.Input(shape=(input_dim,))
        # Encoder



    def get_config(self):
        config = super().get_config()

        encoder_activation_serialized = self.encoder_activation
        if not isinstance(encoder_activation_serialized, str):
            encoder_activation_serialized = keras.layers.serialize(encoder_activation_serialized)

        feature_estimator_activation_serialized = self.feature_estimator_activation
        if not isinstance(feature_estimator_activation_serialized, str):
            feature_estimator_activation_serialized = keras.layers.serialize(feature_estimator_activation_serialized)

        mask_estimator_activation_serialized = self.mask_estimator_activation
        if not isinstance(mask_estimator_activation_serialized, str):
            mask_estimator_activation_serialized = keras.layers.serialize(mask_estimator_activation_serialized)

        config.update({'input_dim': self.input_dim,
                       'p_m': self.p_m,
                       'encoder_activation': self.encoder_activation,
                       'feature_estimator_activation': self.feature_estimator_activation,
                       'mask_estimator_activation': self.mask_estimator_activation,
                       }
                      )
        return config


@keras.saving.register_keras_serializable(package="teras.models")
class VimeSemi(_VimeSemiLF):
    """
    Semi-supervised learning part of the VIME architecture,
    proposed by Jinsung Yoon et al. in the paper
    "VIME: Extending the Success of Self- and
    Semi-supervised Learning to Tabular Domain"

    Reference(s):
        https://proceedings.neurips.cc/paper/2020/hash/7d97667a3e056acab9aaf653807b4a03-Abstract.html

    Args:
        input_dim: ``int``,
            Dimensionality of the input dataset.

        pretrained_encoder: ``keras.Model``,
            Pretrained instance of ``VimEncoder``, pretrained during the
            Self-Supervised learning phase.
            You can access it after training the ``VimeSelf`` model through its
            ``.get_encoder()`` method.

        num_labels: ``int``, default 1,
            Number of labels to predict.

        hidden_dim: ``int``, default 32,
            Dimensionality of hidden layers in the ``VimePredictor`` layer.

        p_m: ``float``, default 0.3,
            Corruption probability

        K: ``int``, default 3,
            Number of augmented samples

        beta: ``float``, default 1.0,
            Hyperparameter to control supervised and unsupervised loss
    """
    def __init__(self,
                 input_dim: int,
                 pretrained_encoder: keras.Model,
                 num_labels: int = 2,
                 hidden_dim: int = 32,
                 p_m: float = 0.3,
                 K: int = 3,
                 beta: float = 1.0,
                 **kwargs):
        mask_and_corrupt = VimeMaskGenerationAndCorruption(p_m,
                                                           name="mask_generation_and_corruption")
        predictor = VimePredictor(num_labels=num_labels,
                                  hidden_dim=hidden_dim,
                                  name="predictor",
                                  )
        super().__init__(input_dim=input_dim,
                         pretrained_encoder=pretrained_encoder,
                         mask_generation_and_corruption=mask_and_corrupt,
                         predictor=predictor,
                         K=K,
                         beta=beta,
                         **kwargs)
        self.input_dim = input_dim
        self.encoder = pretrained_encoder
        self.num_labels = num_labels
        self.hidden_dim = hidden_dim
        self.p_m = p_m
        self.K = K
        self.beta = beta

    def get_config(self):
        config = super().get_config()
        config.update({'input_dim': self.input_dim,
                       'pretrained_encoder': keras.layers.serialize(self.pretrained_encoder),
                       'num_labels': self.num_labels,
                       'hidden_dim': self.hidden_dim,
                       'p_m': self.p_m,
                       'K': self.K,
                       'beta': self.beta,
                       }
                      )
        return config

    @classmethod
    def from_config(cls, config):
        input_dim = config.pop("input_dim")
        pretrained_encoder = keras.layers.deserialize(config.pop("pretrained_encoder"))
        return cls(input_dim=input_dim,
                   pretrained_encoder=pretrained_encoder,
                   **config)



from tensorflow import keras
from teras.utils.types import LayersCollection
from typeguard import check_type


@keras.saving.register_keras_serializable(package="teras.layerflow.models")
class RTDLResNet(keras.Model):
    """
    RTDLResNet model with LayerFlow desing.
    It is based on the ResNet architecture proposed by Yury Gorishniy et al.
    in the paper,
    Revisiting Deep Learning Models for Tabular Data.

    Reference(s):
        https://arxiv.org/abs/2106.11959

    Args:
        input_dim: ``int``,
            The dimensionality of the input dataset,
            or the number of features in the input dataset.

        resnet_blocks: ``List[layers.Layer]`` or ``models.Model`` or ``keras.layers.Layer``,
            List of ``RTDLResNetBlock`` layers to use in the ``RTDLResNet`` model.
            You can pass either pass a list of instances of ``RTDLResNetBlock`` layers,
            or pack them in a Keras model.
            You can import the ``RTDLResNetBlock`` layer as follows,
                >>> from teras.layers import RTDLResNetBlock

        head: ``keras.layers.Layer``,
            An instance of either ``ClassificationHead`` or ``RegressionHead`` layers,
            depending on the task at hand.

            You can import the ``ClassificationHead`` and ``RegressionHead`` layers as follows,
                >>> from teras.layers import ClassificationHead
                >>> from teras.layers import RegressionHead
    """

    def __init__(self,
                 input_dim: int,
                 resnet_blocks: LayersCollection,
                 head: keras.layers.Layer = None,
                 **kwargs):
        # if not isinstance(resnet_blocks, LayersCollection):
        try:
            check_type("resnet_blocks", resnet_blocks, LayersCollection)
        except TypeError:
            raise TypeError("`resnet_blocks` can either be a list of `RTDLResNetBlock` layers "
                            "or a Keras Layer or Model made up of `RTDLResNetBlock` layers. \n"
                            f"But received type: {type(resnet_blocks)}.")
        if isinstance(resnet_blocks, (list, tuple)):
            resnet_blocks = keras.models.Sequential(resnet_blocks,
                                                    name="resnet_blocks")

        inputs = keras.layers.Input(shape=(input_dim,))
        outputs = resnet_blocks(inputs)
        if head is not None:
            outputs = head(outputs)
        super().__init__(inputs=inputs,
                         outputs=outputs,
                         **kwargs)
        self.input_dim = input_dim
        self.resnet_blocks = resnet_blocks
        self.head = head

    def get_config(self):
        config = super().get_config()
        config.update({'input_dim': self.input_dim,
                       'resnet_blocks': keras.layers.serialize(self.resnet_blocks),
                       'head': keras.layers.serialize(self.head)
                       })
        return config

    @classmethod
    def from_config(cls, config):
        input_dim = config.pop("input_dim")
        resnet_blocks = keras.layers.deserialize(config.pop("resnet_blocks"))
        head = keras.layers.deserialize(config.pop("head"))
        return cls(input_dim=input_dim,
                   resnet_blocks=resnet_blocks,
                   head=head,
                   **config)



from teras.config.base import BaseConfig


class TabNetConfig(BaseConfig):
    feature_transformer_dim: int = 32
    decision_step_output_dim: int = 32
    num_decision_steps: int = 5
    num_shared_layers: int = 2
    num_decision_dependent_layers: int = 2
    relaxation_factor: float = 1.5
    batch_momentum: float = 0.9
    virtual_batch_size: int = 64
    residual_normalization_factor: float = 0.5
    epsilon = 1e-5
    categorical_features_vocabulary: dict = None
    encode_categorical_features: bool = True




import tensorflow as tf
from teras.layers.vime.vime_mask_estimator import VimeMaskEstimator


def test_vime_mask_estimator_valid_call():
    inputs = tf.ones((8, 5))
    mask_estimator = VimeMaskEstimator(units=16)
    outputs = mask_estimator(inputs)


def test_vime_mask_estimator_output_shape():
    inputs = tf.ones((8, 5))
    mask_estimator = VimeMaskEstimator(units=16)
    outputs = mask_estimator(inputs)
    assert len(tf.shape(outputs)) == 2
    assert tf.shape(outputs)[0] == 8
    assert tf.shape(outputs)[1] == 16
